from __future__ import annotations

import base64
import json

import pytest
from sqlalchemy import func, select

import app.modules.accounts.service as accounts_service_module
import app.modules.proxy.provider_adapters as provider_adapters_module
from app.core.auth import generate_unique_account_id
from app.core.clients.openai_platform import OpenAIPlatformError, PlatformModelsResponse
from app.core.utils.time import utcnow
from app.db.models import AccountStatus, RequestLog, StickySession, StickySessionKind
from app.db.session import SessionLocal
from app.modules.upstream_identities.repository import (
    OpenAIPlatformIdentitiesRepository,
    OpenAIPlatformIdentityConflictError,
    OpenAIPlatformIdentityCreate,
)

pytestmark = pytest.mark.integration


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str, email: str) -> dict[str, object]:
    payload = {
        "email": email,
        "chatgpt_account_id": account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    return {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access",
            "refreshToken": "refresh",
            "accountId": account_id,
        },
    }


@pytest.mark.asyncio
async def test_import_and_list_accounts(async_client):
    email = "tester@example.com"
    raw_account_id = "acc_explicit"
    payload = {
        "email": email,
        "chatgpt_account_id": "acc_payload",
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    auth_json = {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access",
            "refreshToken": "refresh",
            "accountId": raw_account_id,
        },
    }

    expected_account_id = generate_unique_account_id(raw_account_id, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200
    data = response.json()
    assert data["accountId"] == expected_account_id
    assert data["email"] == email
    assert data["planType"] == "plus"

    list_response = await async_client.get("/api/accounts")
    assert list_response.status_code == 200
    accounts = list_response.json()["accounts"]
    assert any(account["accountId"] == expected_account_id for account in accounts)


@pytest.mark.asyncio
async def test_reactivate_missing_account_returns_404(async_client):
    response = await async_client.post("/api/accounts/missing/reactivate")
    assert response.status_code == 404
    payload = response.json()
    assert payload["error"]["code"] == "account_not_found"


@pytest.mark.asyncio
async def test_pause_missing_account_returns_404(async_client):
    response = await async_client.post("/api/accounts/missing/pause")
    assert response.status_code == 404
    payload = response.json()
    assert payload["error"]["code"] == "account_not_found"


@pytest.mark.asyncio
async def test_pause_account(async_client):
    email = "pause@example.com"
    raw_account_id = "acc_pause"
    payload = {
        "email": email,
        "chatgpt_account_id": raw_account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    auth_json = {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access",
            "refreshToken": "refresh",
            "accountId": raw_account_id,
        },
    }

    expected_account_id = generate_unique_account_id(raw_account_id, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200

    pause = await async_client.post(f"/api/accounts/{expected_account_id}/pause")
    assert pause.status_code == 200
    assert pause.json()["status"] == "paused"

    accounts = await async_client.get("/api/accounts")
    assert accounts.status_code == 200
    data = accounts.json()["accounts"]
    matched = next((account for account in data if account["accountId"] == expected_account_id), None)
    assert matched is not None
    assert matched["status"] == "paused"


@pytest.mark.asyncio
async def test_delete_missing_account_returns_404(async_client):
    response = await async_client.delete("/api/accounts/missing")
    assert response.status_code == 404
    payload = response.json()
    assert payload["error"]["code"] == "account_not_found"


@pytest.mark.asyncio
async def test_list_accounts_includes_platform_request_usage(async_client, monkeypatch):
    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform",
        )

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_usage", "platform-usage@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    create_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Usage",
            "apiKey": "sk-platform-usage",
            "eligibleRouteFamilies": ["public_responses_http"],
        },
    )
    assert create_response.status_code == 200
    identity_id = create_response.json()["accountId"]

    async with SessionLocal() as session:
        session.add(
            RequestLog(
                account_id=None,
                provider_kind="openai_platform",
                routing_subject_id=identity_id,
                api_key_id=None,
                request_id="req_platform_usage",
                requested_at=utcnow(),
                model="gpt-5.1",
                transport="http",
                route_class="openai_public_http",
                upstream_request_id="up_req_platform_usage",
                rejection_reason=None,
                service_tier=None,
                requested_service_tier=None,
                actual_service_tier=None,
                input_tokens=12,
                output_tokens=8,
                cached_input_tokens=3,
                reasoning_tokens=None,
                cost_usd=0.42,
                reasoning_effort=None,
                latency_ms=120,
                latency_first_token_ms=None,
                status="success",
                error_code=None,
                error_message=None,
            )
        )
        await session.commit()

    list_response = await async_client.get("/api/accounts")
    assert list_response.status_code == 200
    platform_account = next(
        account for account in list_response.json()["accounts"] if account["accountId"] == identity_id
    )

    assert platform_account["requestUsage"] == {
        "requestCount": 1,
        "totalTokens": 20,
        "cachedInputTokens": 3,
        "totalCostUsd": 0.42,
    }


@pytest.mark.asyncio
async def test_create_platform_identity_conflict_is_enforced_even_if_service_precheck_is_bypassed(
    async_client,
    monkeypatch,
):
    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform_conflict",
        )

    async def always_allow_platform_creation(self):
        del self
        return False

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )
    monkeypatch.setattr(
        accounts_service_module.AccountsService,
        "_has_platform_identity",
        always_allow_platform_creation,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_conflict_bypass", "platform-conflict-bypass@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    first_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Primary",
            "apiKey": "sk-platform-primary",
            "eligibleRouteFamilies": ["public_models_http"],
        },
    )
    assert first_response.status_code == 200

    second_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Secondary",
            "apiKey": "sk-platform-secondary",
            "eligibleRouteFamilies": ["public_responses_http"],
        },
    )
    assert second_response.status_code == 409
    assert second_response.json()["error"]["code"] == "platform_identity_conflict"


@pytest.mark.asyncio
async def test_update_platform_identity_updates_metadata_without_revalidating_label_only(async_client, monkeypatch):
    create_validation_calls: list[tuple[str, str | None, str | None]] = []
    update_validation_calls: list[tuple[str, str | None, str | None]] = []

    async def fake_create_validate(self, *, api_key, organization=None, project=None):
        del self
        create_validation_calls.append((api_key, organization, project))
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform_update_create",
        )

    async def fail_update_validate(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        update_validation_calls.append(("unexpected", None, None))
        raise AssertionError("label-only edit must not revalidate the platform identity")

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_create_validate,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_edit_label", "platform-edit-label@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    create_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Original",
            "apiKey": "sk-platform-original",
            "organization": "org_original",
            "project": "proj_original",
            "eligibleRouteFamilies": ["public_models_http"],
        },
    )
    assert create_response.status_code == 200
    account_id = create_response.json()["accountId"]
    assert create_validation_calls == [("sk-platform-original", "org_original", "proj_original")]

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fail_update_validate,
    )

    update_response = await async_client.patch(
        f"/api/accounts/platform/{account_id}",
        json={
            "label": "Platform Renamed",
            "eligibleRouteFamilies": ["public_models_http", "public_responses_http"],
        },
    )
    assert update_response.status_code == 200
    payload = update_response.json()
    assert payload["accountId"] == account_id
    assert payload["label"] == "Platform Renamed"
    assert payload["organization"] == "org_original"
    assert payload["project"] == "proj_original"
    assert sorted(payload["eligibleRouteFamilies"]) == ["public_models_http", "public_responses_http"]
    assert update_validation_calls == []


@pytest.mark.asyncio
async def test_update_platform_identity_revalidates_auth_affecting_fields(async_client, monkeypatch):
    validation_calls: list[tuple[str, str | None, str | None]] = []

    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self
        validation_calls.append((api_key, organization, project))
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform_update_auth",
        )

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_edit_auth", "platform-edit-auth@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    create_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Original",
            "apiKey": "sk-platform-original",
            "organization": "org_original",
            "project": "proj_original",
            "eligibleRouteFamilies": ["public_models_http"],
        },
    )
    assert create_response.status_code == 200
    account_id = create_response.json()["accountId"]

    update_response = await async_client.patch(
        f"/api/accounts/platform/{account_id}",
        json={
            "organization": "org_updated",
            "project": None,
        },
    )
    assert update_response.status_code == 200
    payload = update_response.json()
    assert payload["accountId"] == account_id
    assert payload["organization"] == "org_updated"
    assert payload["project"] is None
    assert validation_calls == [
        ("sk-platform-original", "org_original", "proj_original"),
        ("sk-platform-original", "org_updated", None),
    ]


@pytest.mark.asyncio
async def test_update_platform_identity_revalidates_key_rotation(async_client, monkeypatch):
    validation_calls: list[tuple[str, str | None, str | None]] = []

    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self
        validation_calls.append((api_key, organization, project))
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform_rotate",
        )

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_rotate", "platform-rotate@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    create_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Original",
            "apiKey": "sk-platform-original",
            "organization": "org_original",
            "project": "proj_original",
            "eligibleRouteFamilies": ["public_models_http"],
        },
    )
    assert create_response.status_code == 200
    account_id = create_response.json()["accountId"]

    update_response = await async_client.patch(
        f"/api/accounts/platform/{account_id}",
        json={"apiKey": "sk-platform-rotated"},
    )
    assert update_response.status_code == 200
    assert validation_calls == [
        ("sk-platform-original", "org_original", "proj_original"),
        ("sk-platform-rotated", "org_original", "proj_original"),
    ]


@pytest.mark.asyncio
async def test_update_platform_identity_surfaces_auth_failure_reason_for_rotated_key(async_client, monkeypatch):
    async def create_validate(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform_auth_failure_create",
        )

    async def fail_update_validate(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        raise OpenAIPlatformError(
            401,
            {
                "error": {
                    "code": "invalid_api_key",
                    "message": "The supplied API key is invalid.",
                    "type": "invalid_request_error",
                }
            },
        )

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        create_validate,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_auth_fail", "platform-auth-fail@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    create_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Auth Failure",
            "apiKey": "sk-platform-original",
            "eligibleRouteFamilies": ["public_models_http"],
        },
    )
    assert create_response.status_code == 200
    account_id = create_response.json()["accountId"]

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fail_update_validate,
    )

    update_response = await async_client.patch(
        f"/api/accounts/platform/{account_id}",
        json={"apiKey": "sk-platform-invalid"},
    )
    assert update_response.status_code == 200
    payload = update_response.json()
    assert payload["status"] == "deactivated"
    assert payload["lastAuthFailureReason"] == "The supplied API key is invalid."

    async with SessionLocal() as session:
        repo = OpenAIPlatformIdentitiesRepository(session)
        identity = await repo.get_by_id(account_id)
        assert identity is not None
        decrypting_service = accounts_service_module.OpenAIPlatformIdentitiesService(repo)
        assert decrypting_service.decrypt_api_key(identity) == "sk-platform-invalid"
        assert identity.last_auth_failure_reason == "The supplied API key is invalid."
        assert identity.status == AccountStatus.DEACTIVATED


@pytest.mark.asyncio
async def test_delete_platform_identity_cleans_up_provider_scoped_sticky_sessions(async_client, monkeypatch):
    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform_delete",
        )

    monkeypatch.setattr(
        provider_adapters_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_delete", "platform-delete@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    create_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Delete",
            "apiKey": "sk-platform-delete",
            "eligibleRouteFamilies": ["public_responses_http"],
        },
    )
    assert create_response.status_code == 200
    identity_id = create_response.json()["accountId"]

    async with SessionLocal() as session:
        session.add(
            StickySession(
                key="platform-sticky",
                kind=StickySessionKind.PROMPT_CACHE,
                provider_kind="openai_platform",
                account_id=None,
                routing_subject_id=identity_id,
            )
        )
        await session.commit()

    delete_response = await async_client.delete(f"/api/accounts/{identity_id}")
    assert delete_response.status_code == 200

    async with SessionLocal() as session:
        remaining = await session.scalar(
            select(func.count())
            .select_from(StickySession)
            .where(
                StickySession.provider_kind == "openai_platform",
                StickySession.routing_subject_id == identity_id,
            )
        )
        assert remaining == 0


@pytest.mark.asyncio
async def test_update_platform_identity_returns_404_for_missing_identity(async_client):
    response = await async_client.patch(
        "/api/accounts/platform/missing-platform",
        json={"label": "Missing Platform"},
    )
    assert response.status_code == 404
    assert response.json()["error"]["code"] == "account_not_found"


@pytest.mark.asyncio
async def test_update_platform_identity_returns_404_for_chatgpt_account(async_client):
    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_edit_mismatch", "platform-edit-mismatch@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200
    account_id = import_response.json()["accountId"]

    response = await async_client.patch(
        f"/api/accounts/platform/{account_id}",
        json={"label": "Should Fail"},
    )
    assert response.status_code == 404
    assert response.json()["error"]["code"] == "account_not_found"


@pytest.mark.asyncio
async def test_update_platform_identity_rejects_invalid_route_family(async_client, monkeypatch):
    async def fake_validate_platform_identity(self, *, api_key, organization=None, project=None):
        del self, api_key, organization, project
        return PlatformModelsResponse(
            payload={"object": "list", "data": [{"id": "gpt-5.1", "object": "model", "owned_by": "openai"}]},
            upstream_request_id="up_req_validate_platform_update_invalid",
        )

    monkeypatch.setattr(
        accounts_service_module.OpenAIPlatformProviderAdapter,
        "validate_identity",
        fake_validate_platform_identity,
    )

    import_response = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_platform_edit_invalid", "platform-edit-invalid@example.com")),
                "application/json",
            )
        },
    )
    assert import_response.status_code == 200

    create_response = await async_client.post(
        "/api/accounts/platform",
        json={
            "label": "Platform Original",
            "apiKey": "sk-platform-original",
            "eligibleRouteFamilies": ["public_models_http"],
        },
    )
    assert create_response.status_code == 200
    account_id = create_response.json()["accountId"]

    update_response = await async_client.patch(
        f"/api/accounts/platform/{account_id}",
        json={"eligibleRouteFamilies": ["not_a_real_route_family"]},
    )
    assert update_response.status_code == 422


@pytest.mark.asyncio
async def test_platform_identity_repository_enforces_singleton(async_client):
    del async_client
    async with SessionLocal() as session:
        repo = OpenAIPlatformIdentitiesRepository(session)
        first = OpenAIPlatformIdentityCreate(
            id="platform_one",
            label="Platform One",
            api_key_encrypted=b"encrypted-1",
            organization_id=None,
            project_id=None,
            eligible_route_families=(),
            status=AccountStatus.ACTIVE,
            last_validated_at=None,
            last_auth_failure_reason=None,
        )
        second = OpenAIPlatformIdentityCreate(
            id="platform_two",
            label="Platform Two",
            api_key_encrypted=b"encrypted-2",
            organization_id=None,
            project_id=None,
            eligible_route_families=(),
            status=AccountStatus.ACTIVE,
            last_validated_at=None,
            last_auth_failure_reason=None,
        )

        await repo.create_identity(first)
        with pytest.raises(OpenAIPlatformIdentityConflictError):
            await repo.create_identity(second)



"""add persisted cost to request_logs

Revision ID: 20260325_000000_add_request_log_cost
Revises: 20260321_210000_merge_request_log_tiers_and_dashboard_index_heads
Create Date: 2026-03-25
"""

from __future__ import annotations

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

from app.core.usage.pricing import UsageTokens, calculate_cost_from_usage, get_pricing_for_model

# revision identifiers, used by Alembic.
revision = "20260325_000000_add_request_log_cost"
down_revision = "20260321_210000_merge_request_log_tiers_and_dashboard_index_heads"
branch_labels = None
depends_on = None

_BACKFILL_BATCH_SIZE = 1000
_TEMP_COST_TABLE_NAME = "request_log_cost_backfill"


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {column["name"] for column in inspector.get_columns(table_name)}


def _calculate_cost(
    *,
    model: str | None,
    service_tier: str | None,
    input_tokens: int | None,
    output_tokens: int | None,
    cached_input_tokens: int | None,
    reasoning_tokens: int | None,
) -> float | None:
    if not model or input_tokens is None:
        return None
    resolved_output_tokens = output_tokens if output_tokens is not None else reasoning_tokens
    if resolved_output_tokens is None:
        return None
    resolved = get_pricing_for_model(model, None, None)
    if resolved is None:
        return None
    _, price = resolved
    normalized_cached_tokens = max(0, min(int(cached_input_tokens or 0), int(input_tokens)))
    return calculate_cost_from_usage(
        UsageTokens(
            input_tokens=float(input_tokens),
            output_tokens=float(resolved_output_tokens),
            cached_input_tokens=float(normalized_cached_tokens),
        ),
        price,
        service_tier=service_tier,
    )


def _create_backfill_temp_table(bind: Connection) -> None:
    bind.execute(sa.text(f"DROP TABLE IF EXISTS {_TEMP_COST_TABLE_NAME}"))
    bind.execute(
        sa.text(
            f"""
            CREATE TEMPORARY TABLE {_TEMP_COST_TABLE_NAME} (
                id INTEGER PRIMARY KEY,
                cost_usd FLOAT NULL
            )
            """
        )
    )


def _apply_cost_backfill_batch(
    bind: Connection,
    batch: Sequence[dict[str, int | float | None]],
) -> None:
    if not batch:
        return

    bind.execute(sa.text(f"DELETE FROM {_TEMP_COST_TABLE_NAME}"))
    bind.execute(
        sa.text(f"INSERT INTO {_TEMP_COST_TABLE_NAME} (id, cost_usd) VALUES (:id, :cost_usd)"),
        list(batch),
    )
    if bind.dialect.name == "sqlite":
        bind.execute(
            sa.text(
                f"""
                UPDATE request_logs
                SET cost_usd = (
                    SELECT tmp.cost_usd
                    FROM {_TEMP_COST_TABLE_NAME} AS tmp
                    WHERE tmp.id = request_logs.id
                )
                WHERE id IN (SELECT id FROM {_TEMP_COST_TABLE_NAME})
                """
            )
        )
        return

    bind.execute(
        sa.text(
            f"""
            UPDATE request_logs
            SET cost_usd = tmp.cost_usd
            FROM {_TEMP_COST_TABLE_NAME} AS tmp
            WHERE request_logs.id = tmp.id
            """
        )
    )


def upgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "request_logs")
    if not columns:
        return

    if "cost_usd" not in columns:
        with op.batch_alter_table("request_logs") as batch_op:
            batch_op.add_column(sa.Column("cost_usd", sa.Float(), nullable=True))

    request_logs = sa.table(
        "request_logs",
        sa.column("id", sa.Integer()),
        sa.column("model", sa.String()),
        sa.column("service_tier", sa.String()),
        sa.column("input_tokens", sa.Integer()),
        sa.column("output_tokens", sa.Integer()),
        sa.column("cached_input_tokens", sa.Integer()),
        sa.column("reasoning_tokens", sa.Integer()),
        sa.column("cost_usd", sa.Float()),
    )

    last_seen_id = 0
    _create_backfill_temp_table(bind)
    try:
        while True:
            rows = (
                bind.execute(
                    sa.select(
                        request_logs.c.id,
                        request_logs.c.model,
                        request_logs.c.service_tier,
                        request_logs.c.input_tokens,
                        request_logs.c.output_tokens,
                        request_logs.c.cached_input_tokens,
                        request_logs.c.reasoning_tokens,
                    )
                    .where(request_logs.c.id > last_seen_id)
                    .order_by(request_logs.c.id)
                    .limit(_BACKFILL_BATCH_SIZE)
                )
                .mappings()
                .all()
            )
            if not rows:
                break

            batch = [
                {
                    "id": row["id"],
                    "cost_usd": _calculate_cost(
                        model=row["model"],
                        service_tier=row["service_tier"],
                        input_tokens=row["input_tokens"],
                        output_tokens=row["output_tokens"],
                        cached_input_tokens=row["cached_input_tokens"],
                        reasoning_tokens=row["reasoning_tokens"],
                    ),
                }
                for row in rows
            ]
            _apply_cost_backfill_batch(bind, batch)
            last_seen_id = int(rows[-1]["id"])
    finally:
        bind.execute(sa.text(f"DROP TABLE IF EXISTS {_TEMP_COST_TABLE_NAME}"))


def downgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "request_logs")
    if not columns or "cost_usd" not in columns:
        return

    with op.batch_alter_table("request_logs") as batch_op:
        batch_op.drop_column("cost_usd")



"""make request_logs.account_id nullable for no-account errors

Revision ID: 20260309_000000_request_logs_nullable_account_id
Revises: 20260308_000000_add_sqlite_performance_indexes
Create Date: 2026-03-09
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "20260309_000000_request_logs_nullable_account_id"
down_revision = "20260308_000000_add_sqlite_performance_indexes"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("request_logs") as batch_op:
        batch_op.alter_column("account_id", nullable=True)


def downgrade() -> None:
    op.execute(sa.text("DELETE FROM request_logs WHERE account_id IS NULL"))
    with op.batch_alter_table("request_logs") as batch_op:
        batch_op.alter_column("account_id", nullable=False)



from __future__ import annotations

import json

import pytest

import app.modules.dashboard_auth.service as dashboard_auth_service_module
from app.core.crypto import TokenEncryptor
from app.modules.dashboard_auth.service import DashboardSessionStore

pytestmark = pytest.mark.unit


def test_session_store_round_trip_with_password_and_totp(monkeypatch) -> None:
    monkeypatch.setattr(dashboard_auth_service_module, "time", lambda: 1_700_000_000)
    store = DashboardSessionStore()

    session_id = store.create(password_verified=True, totp_verified=False)
    state = store.get(session_id)

    assert state is not None
    assert state.password_verified is True
    assert state.totp_verified is False
    assert state.expires_at == 1_700_000_000 + 12 * 60 * 60


def test_session_store_rejects_legacy_cookie_without_password_flag(monkeypatch) -> None:
    monkeypatch.setattr(dashboard_auth_service_module, "time", lambda: 1_700_000_000)
    store = DashboardSessionStore()
    encryptor = TokenEncryptor()
    legacy_payload = json.dumps({"exp": 1_700_000_100, "tv": True}, separators=(",", ":"))
    legacy_session_id = encryptor.encrypt(legacy_payload).decode("ascii")

    assert store.get(legacy_session_id) is None


def test_session_store_rejects_expired_session(monkeypatch) -> None:
    current = {"value": 1_700_000_000}
    monkeypatch.setattr(dashboard_auth_service_module, "time", lambda: current["value"])
    store = DashboardSessionStore()

    session_id = store.create(password_verified=True, totp_verified=True)
    current["value"] += 12 * 60 * 60 + 1

    assert store.get(session_id) is None



from __future__ import annotations

from typing import Literal, TypeAlias

CHATGPT_WEB_PROVIDER_KIND = "chatgpt_web"
OPENAI_PLATFORM_PROVIDER_KIND = "openai_platform"

PUBLIC_MODELS_HTTP_ROUTE_FAMILY = "public_models_http"
PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY = "public_responses_http"
BACKEND_CODEX_HTTP_ROUTE_FAMILY = "backend_codex_http"

OPENAI_PUBLIC_HTTP_ROUTE_CLASS = "openai_public_http"
OPENAI_PUBLIC_WS_ROUTE_CLASS = "openai_public_ws"
CHATGPT_PRIVATE_ROUTE_CLASS = "chatgpt_private"

ProviderKind: TypeAlias = Literal["chatgpt_web", "openai_platform"]
PlatformRouteFamily: TypeAlias = Literal["public_models_http", "public_responses_http", "backend_codex_http"]

PHASE1_PLATFORM_ROUTE_FAMILIES: tuple[PlatformRouteFamily, ...] = (
    PUBLIC_MODELS_HTTP_ROUTE_FAMILY,
    PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
    BACKEND_CODEX_HTTP_ROUTE_FAMILY,
)

SUPPORTED_PROVIDER_KINDS: tuple[ProviderKind, ...] = (
    CHATGPT_WEB_PROVIDER_KIND,
    OPENAI_PLATFORM_PROVIDER_KIND,
)



from __future__ import annotations

import hashlib
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from datetime import UTC, datetime
from types import SimpleNamespace
from typing import cast

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient

import app.core.auth.dependencies as auth_dependencies
import app.core.middleware.api_firewall as api_firewall_module
from app.core.auth.api_key_cache import get_api_key_cache
from app.core.crypto import TokenEncryptor
from app.core.middleware.api_firewall import add_api_firewall_middleware
from app.core.middleware.firewall_cache import get_firewall_ip_cache
from app.db.models import Account, AccountStatus, UsageHistory
from app.modules.api_keys.service import ApiKeyData
from app.modules.proxy.account_cache import get_account_selection_cache
from app.modules.proxy.load_balancer import LoadBalancer
from app.modules.proxy.repo_bundle import ProxyRepoFactory

pytestmark = pytest.mark.unit


@pytest.fixture(autouse=True)
def _clear_hot_path_caches() -> None:
    get_api_key_cache().clear()
    get_firewall_ip_cache().invalidate_all()
    get_account_selection_cache().invalidate()


@pytest.mark.asyncio
async def test_api_key_validation_uses_cache_for_repeated_key(monkeypatch: pytest.MonkeyPatch) -> None:
    api_key_data = ApiKeyData(
        id="key_1",
        name="hot-path",
        key_prefix="sk-clb-test",
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=True,
        created_at=datetime.now(UTC),
        last_used_at=None,
    )
    calls = 0

    class _SettingsCache:
        async def get(self) -> SimpleNamespace:
            return SimpleNamespace(api_key_auth_enabled=True)

    class _Service:
        def __init__(self, _repo: object) -> None:
            pass

        async def validate_key(self, _token: str) -> ApiKeyData:
            nonlocal calls
            calls += 1
            return api_key_data

    @asynccontextmanager
    async def _fake_session() -> AsyncIterator[object]:
        yield object()

    monkeypatch.setattr(auth_dependencies, "get_settings_cache", lambda: _SettingsCache())
    monkeypatch.setattr(auth_dependencies, "get_background_session", _fake_session)
    monkeypatch.setattr(auth_dependencies, "ApiKeysRepository", lambda _session: object())
    monkeypatch.setattr(auth_dependencies, "ApiKeysService", _Service)

    token = "sk-clb-hot-path"
    expected_hash = hashlib.sha256(token.encode("utf-8")).hexdigest()

    for _ in range(10):
        resolved = await auth_dependencies.validate_proxy_api_key_authorization(f"Bearer {token}")
        assert resolved == api_key_data

    assert calls == 1
    cache = get_api_key_cache()
    assert expected_hash in cache._cache
    assert token not in cache._cache


@pytest.mark.asyncio
async def test_firewall_middleware_uses_cache_for_repeated_ip(monkeypatch: pytest.MonkeyPatch) -> None:
    calls = 0

    class _Service:
        def __init__(self, _repo: object) -> None:
            pass

        async def is_ip_allowed(self, _ip: str | None) -> bool:
            nonlocal calls
            calls += 1
            return True

    @asynccontextmanager
    async def _fake_session() -> AsyncIterator[object]:
        yield object()

    monkeypatch.setattr(
        api_firewall_module,
        "get_settings",
        lambda: SimpleNamespace(firewall_trusted_proxy_cidrs=[], firewall_trust_proxy_headers=False),
    )
    monkeypatch.setattr(api_firewall_module, "get_background_session", _fake_session)
    monkeypatch.setattr(api_firewall_module, "FirewallRepository", lambda _session: object())
    monkeypatch.setattr(api_firewall_module, "FirewallService", _Service)

    app = FastAPI()
    add_api_firewall_middleware(app)

    @app.get("/v1/test")
    async def _v1_test() -> dict[str, str]:
        return {"ok": "true"}

    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://testserver") as client:
        for _ in range(10):
            response = await client.get("/v1/test")
            assert response.status_code == 200

    assert calls == 1


@pytest.mark.asyncio
async def test_account_selection_cache_reuses_inputs_and_invalidates_on_refresh() -> None:
    encryptor = TokenEncryptor()
    now = datetime.now(UTC)
    now_epoch = int(now.timestamp())
    account = Account(
        id="acc-hot-path",
        chatgpt_account_id="workspace-acc-hot-path",
        email="hot-path@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )
    primary = UsageHistory(
        id=1,
        account_id=account.id,
        recorded_at=now,
        window="primary",
        used_percent=15.0,
        reset_at=now_epoch + 300,
        window_minutes=5,
    )
    secondary = UsageHistory(
        id=2,
        account_id=account.id,
        recorded_at=now,
        window="secondary",
        used_percent=25.0,
        reset_at=now_epoch + 1800,
        window_minutes=30,
    )

    class _AccountsRepo:
        def __init__(self) -> None:
            self.calls = 0

        async def list_accounts(self) -> list[Account]:
            self.calls += 1
            return [account]

    class _UsageRepo:
        def __init__(self) -> None:
            self.primary_calls = 0
            self.secondary_calls = 0

        async def latest_by_account(self, window: str | None = None) -> dict[str, UsageHistory]:
            if window == "secondary":
                self.secondary_calls += 1
                return {account.id: secondary}
            self.primary_calls += 1
            return {account.id: primary}

    class _Repos:
        def __init__(self, accounts_repo: _AccountsRepo, usage_repo: _UsageRepo) -> None:
            self.accounts = accounts_repo
            self.usage = usage_repo

        async def __aenter__(self) -> "_Repos":
            return self

        async def __aexit__(self, exc_type, exc, tb) -> None:
            return None

    accounts_repo = _AccountsRepo()
    usage_repo = _UsageRepo()

    def _repo_factory() -> _Repos:
        return _Repos(accounts_repo, usage_repo)

    # Create cache with explicit TTL > 0 (not the global singleton which may be disabled in tests)
    from app.modules.proxy.account_cache import AccountSelectionCache

    cache = AccountSelectionCache(ttl_seconds=5)
    balancer = LoadBalancer(cast(ProxyRepoFactory, _repo_factory))
    balancer._selection_inputs_cache = cache  # Override with test-specific cache

    for _ in range(10):
        inputs = await balancer._load_selection_inputs(model=None)
        assert len(inputs.accounts) == 1
        assert inputs.accounts[0].id == account.id

    assert accounts_repo.calls == 1
    assert usage_repo.primary_calls == 1
    assert usage_repo.secondary_calls == 1

    cache.invalidate()
    refreshed = await balancer._load_selection_inputs(model=None)
    assert len(refreshed.accounts) == 1
    assert refreshed.accounts[0].id == account.id
    assert accounts_repo.calls == 2
    assert usage_repo.primary_calls == 2
    assert usage_repo.secondary_calls == 2


# ---------------------------------------------------------------------------
# Task 1 RED: API key cache invalidation (xfail — bug not yet fixed)
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_deleted_key_rejected_immediately() -> None:
    """BUG: ApiKeyCache is never invalidated when a key is deleted."""
    plain_key = "sk-clb-test-del-0001"
    key_hash = hashlib.sha256(plain_key.encode()).hexdigest()
    now = datetime.now(UTC)
    api_key_data = ApiKeyData(
        id="key-del-1",
        name="test-delete",
        key_prefix=plain_key[:15],
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=True,
        created_at=now,
        last_used_at=None,
    )

    class _DeleteOnlyRepo:
        async def get_by_id(self, _key_id: str) -> SimpleNamespace:
            return SimpleNamespace(key_hash=key_hash)

        async def delete(self, _key_id: str) -> bool:
            return True

    cache = get_api_key_cache()
    await cache.set(key_hash, api_key_data)

    from app.modules.api_keys.service import ApiKeysService

    service = ApiKeysService(_DeleteOnlyRepo())  # type: ignore[arg-type]
    await service.delete_key("key-del-1")

    # BUG: after deletion the cache should be empty, but it still holds stale data
    cached = await cache.get(key_hash)
    assert cached is None  # xfail: cache still returns the deleted key's data


@pytest.mark.asyncio
async def test_regenerated_key_old_token_rejected_immediately() -> None:
    """BUG: Old key hash stays in cache after regeneration — old token remains valid."""
    plain_key = "sk-clb-test-regen-001"
    old_key_hash = hashlib.sha256(plain_key.encode()).hexdigest()
    now = datetime.now(UTC)
    api_key_data = ApiKeyData(
        id="key-regen-1",
        name="test-regen",
        key_prefix=plain_key[:15],
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=True,
        created_at=now,
        last_used_at=None,
    )
    old_row = SimpleNamespace(
        id="key-regen-1",
        name="test-regen",
        key_hash=old_key_hash,
        key_prefix=plain_key[:15],
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=True,
        created_at=now,
        last_used_at=None,
        limits=[],
    )
    new_plain = "sk-clb-new-key-regen"
    new_row = SimpleNamespace(
        id="key-regen-1",
        name="test-regen",
        key_hash=hashlib.sha256(new_plain.encode()).hexdigest(),
        key_prefix=new_plain[:15],
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=True,
        created_at=now,
        last_used_at=None,
        limits=[],
    )

    class _RegenRepo:
        async def get_by_id(self, _key_id: str) -> SimpleNamespace:
            return old_row

        async def update(self, _key_id: str, **_kwargs: object) -> SimpleNamespace:
            return new_row

    cache = get_api_key_cache()
    await cache.set(old_key_hash, api_key_data)

    from app.modules.api_keys.service import ApiKeysService

    service = ApiKeysService(_RegenRepo())  # type: ignore[arg-type]
    await service.regenerate_key("key-regen-1")

    # BUG: old token should be evicted from cache but it still authorises requests
    cached = await cache.get(old_key_hash)
    assert cached is None  # xfail: old hash still in cache


@pytest.mark.asyncio
async def test_deactivated_key_rejected_immediately() -> None:
    """BUG: Deactivated key (is_active=False) stays in cache — still grants access."""
    plain_key = "sk-clb-test-deact-01"
    key_hash = hashlib.sha256(plain_key.encode()).hexdigest()
    now = datetime.now(UTC)
    api_key_data = ApiKeyData(
        id="key-deact-1",
        name="test-deactivate",
        key_prefix=plain_key[:15],
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=True,
        created_at=now,
        last_used_at=None,
    )
    deactivated_row = SimpleNamespace(
        id="key-deact-1",
        name="test-deactivate",
        key_hash=key_hash,
        key_prefix=plain_key[:15],
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=False,
        created_at=now,
        last_used_at=None,
        limits=[],
    )

    class _UpdateOnlyRepo:
        async def get_by_id(self, _key_id: str) -> SimpleNamespace:
            return deactivated_row

        async def update(self, _key_id: str, **_kwargs: object) -> SimpleNamespace:
            return deactivated_row

        async def commit(self) -> None:
            return None

        async def rollback(self) -> None:
            return None

    cache = get_api_key_cache()
    await cache.set(key_hash, api_key_data)

    from app.modules.api_keys.service import ApiKeysService, ApiKeyUpdateData

    service = ApiKeysService(_UpdateOnlyRepo())  # type: ignore[arg-type]
    await service.update_key("key-deact-1", ApiKeyUpdateData(is_active=False, is_active_set=True))

    # BUG: deactivated key should be evicted from cache but it still returns stale active data
    cached = await cache.get(key_hash)
    assert cached is None  # xfail: cache still returns the deactivated key's data


# ---------------------------------------------------------------------------
# Task 3 RED: selection cache poisoning (xfail — bug not yet fixed)
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_different_models_get_different_cache_entries() -> None:
    """BUG: First model query caches its result; second model query gets wrong cached data."""
    from app.modules.proxy.account_cache import AccountSelectionCache

    cache = AccountSelectionCache(ttl_seconds=5)

    class _AccountsRepo:
        def __init__(self) -> None:
            self.calls = 0

        async def list_accounts(self) -> list[object]:
            self.calls += 1
            return []

    class _UsageRepo:
        async def latest_by_account(self, window: str | None = None) -> dict[object, object]:
            return {}

    class _Repos:
        def __init__(self, accounts_repo: _AccountsRepo, usage_repo: _UsageRepo) -> None:
            self.accounts = accounts_repo
            self.usage = usage_repo

        async def __aenter__(self) -> "_Repos":
            return self

        async def __aexit__(self, exc_type: object, exc: object, tb: object) -> None:
            return None

    accounts_repo = _AccountsRepo()
    usage_repo = _UsageRepo()

    def _repo_factory() -> _Repos:
        return _Repos(accounts_repo, usage_repo)

    balancer = LoadBalancer(cast(ProxyRepoFactory, _repo_factory))
    balancer._selection_inputs_cache = cache

    # First call for model=None — loads from DB and caches
    await balancer._load_selection_inputs(model=None)
    assert accounts_repo.calls == 1

    # Second call for a different model — should load from DB again (different cache key)
    # BUG: it hits the single-slot cache and skips DB load
    await balancer._load_selection_inputs(model="gpt-4-no-filter")
    assert accounts_repo.calls == 2  # xfail: still 1 because cache poisoned this call


@pytest.mark.asyncio
async def test_same_model_reuses_cache() -> None:
    """Same (model, limit_name) pair must reuse the cached result."""
    from app.modules.proxy.account_cache import AccountSelectionCache

    cache = AccountSelectionCache(ttl_seconds=5)

    class _AccountsRepo:
        def __init__(self) -> None:
            self.calls = 0

        async def list_accounts(self) -> list[object]:
            self.calls += 1
            return []

    class _UsageRepo:
        async def latest_by_account(self, window: str | None = None) -> dict[object, object]:
            return {}

    class _Repos:
        def __init__(self, accounts_repo: _AccountsRepo, usage_repo: _UsageRepo) -> None:
            self.accounts = accounts_repo
            self.usage = usage_repo

        async def __aenter__(self) -> "_Repos":
            return self

        async def __aexit__(self, exc_type: object, exc: object, tb: object) -> None:
            return None

    accounts_repo = _AccountsRepo()
    usage_repo = _UsageRepo()

    def _repo_factory() -> _Repos:
        return _Repos(accounts_repo, usage_repo)

    balancer = LoadBalancer(cast(ProxyRepoFactory, _repo_factory))
    balancer._selection_inputs_cache = cache

    # Two calls with same model — second should be served from cache
    await balancer._load_selection_inputs(model=None)
    await balancer._load_selection_inputs(model=None)
    assert accounts_repo.calls == 1  # second call must be a cache hit


@pytest.mark.asyncio
async def test_different_limit_names_get_different_cache_entries() -> None:
    """BUG: Queries with different additional_limit_name share the same cache slot."""
    from app.modules.proxy.account_cache import AccountSelectionCache

    cache = AccountSelectionCache(ttl_seconds=5)

    class _AccountsRepo:
        def __init__(self) -> None:
            self.calls = 0

        async def list_accounts(self) -> list[object]:
            self.calls += 1
            return []

    class _UsageRepo:
        async def latest_by_account(self, window: str | None = None) -> dict[object, object]:
            return {}

    class _Repos:
        def __init__(self, accounts_repo: _AccountsRepo, usage_repo: _UsageRepo) -> None:
            self.accounts = accounts_repo
            self.usage = usage_repo

        async def __aenter__(self) -> "_Repos":
            return self

        async def __aexit__(self, exc_type: object, exc: object, tb: object) -> None:
            return None

    accounts_repo = _AccountsRepo()
    usage_repo = _UsageRepo()

    def _repo_factory() -> _Repos:
        return _Repos(accounts_repo, usage_repo)

    balancer = LoadBalancer(cast(ProxyRepoFactory, _repo_factory))
    balancer._selection_inputs_cache = cache

    # First call with limit_name=None — loads and caches
    await balancer._load_selection_inputs(model=None, additional_limit_name=None)
    assert accounts_repo.calls == 1

    # Second call with different additional_limit_name — should re-load from DB
    # BUG: same single slot used → DB never re-queried
    await balancer._load_selection_inputs(model=None, additional_limit_name="pro_tier")
    assert accounts_repo.calls == 2  # xfail: still 1 because of cache poisoning


@pytest.mark.asyncio
async def test_invalidate_clears_all_keyed_entries() -> None:
    """invalidate() must clear all cached entries (single-slot or multi-key)."""
    from app.modules.proxy.account_cache import AccountSelectionCache

    cache = AccountSelectionCache(ttl_seconds=5)

    class _AccountsRepo:
        def __init__(self) -> None:
            self.calls = 0

        async def list_accounts(self) -> list[object]:
            self.calls += 1
            return []

    class _UsageRepo:
        async def latest_by_account(self, window: str | None = None) -> dict[object, object]:
            return {}

    class _Repos:
        def __init__(self, accounts_repo: _AccountsRepo, usage_repo: _UsageRepo) -> None:
            self.accounts = accounts_repo
            self.usage = usage_repo

        async def __aenter__(self) -> "_Repos":
            return self

        async def __aexit__(self, exc_type: object, exc: object, tb: object) -> None:
            return None

    accounts_repo = _AccountsRepo()
    usage_repo = _UsageRepo()

    def _repo_factory() -> _Repos:
        return _Repos(accounts_repo, usage_repo)

    balancer = LoadBalancer(cast(ProxyRepoFactory, _repo_factory))
    balancer._selection_inputs_cache = cache

    # Populate cache
    await balancer._load_selection_inputs(model=None)
    assert accounts_repo.calls == 1

    # Invalidate should clear everything
    cache.invalidate()

    # Next call must re-load from DB
    await balancer._load_selection_inputs(model=None)
    assert accounts_repo.calls == 2  # must re-load after invalidation



from __future__ import annotations

from datetime import timedelta

import pytest
from sqlalchemy import update

from app.core.crypto import TokenEncryptor
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus, ApiKey
from app.db.session import SessionLocal
from app.modules.accounts.repository import AccountsRepository
from app.modules.request_logs.repository import RequestLogsRepository

pytestmark = pytest.mark.integration


def _make_account(account_id: str, email: str) -> Account:
    encryptor = TokenEncryptor()
    return Account(
        id=account_id,
        email=email,
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=utcnow(),
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )


def _cost(
    input_tokens: int,
    output_tokens: int,
    cached_tokens: int = 0,
    *,
    input_rate: float = 1.25,
    cached_rate: float = 0.125,
    output_rate: float = 10.0,
) -> float:
    billable = input_tokens - cached_tokens
    return (
        (billable / 1_000_000) * input_rate
        + (cached_tokens / 1_000_000) * cached_rate
        + (output_tokens / 1_000_000) * output_rate
    )


@pytest.mark.asyncio
async def test_request_logs_status_ok_filters_success(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_ok", "ok@example.com"))

        await logs_repo.add_log(
            account_id="acc_ok",
            request_id="req_ok_1",
            model="gpt-5.1",
            input_tokens=10,
            output_tokens=20,
            latency_ms=100,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=2),
        )
        await logs_repo.add_log(
            account_id="acc_ok",
            request_id="req_ok_2",
            model="gpt-5.1",
            input_tokens=5,
            output_tokens=0,
            latency_ms=50,
            status="error",
            error_code="rate_limit_exceeded",
            requested_at=now - timedelta(minutes=1),
        )

    response = await async_client.get("/api/request-logs?status=ok")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    assert payload[0]["status"] == "ok"
    assert payload[0]["errorCode"] is None


@pytest.mark.asyncio
async def test_request_logs_status_rate_limit_filters_codes(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_rate", "rate@example.com"))

        await logs_repo.add_log(
            account_id="acc_rate",
            request_id="req_rate_1",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="rate_limit_exceeded",
            requested_at=now - timedelta(minutes=1),
        )
        await logs_repo.add_log(
            account_id="acc_rate",
            request_id="req_rate_2",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="insufficient_quota",
            requested_at=now - timedelta(minutes=2),
        )

    response = await async_client.get("/api/request-logs?status=rate_limit")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    assert payload[0]["status"] == "rate_limit"
    assert payload[0]["errorCode"] == "rate_limit_exceeded"


@pytest.mark.asyncio
async def test_request_logs_status_quota_filters_codes(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_quota", "quota@example.com"))

        await logs_repo.add_log(
            account_id="acc_quota",
            request_id="req_quota_1",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="insufficient_quota",
            requested_at=now - timedelta(minutes=3),
        )
        await logs_repo.add_log(
            account_id="acc_quota",
            request_id="req_quota_2",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="usage_not_included",
            requested_at=now - timedelta(minutes=2),
        )
        await logs_repo.add_log(
            account_id="acc_quota",
            request_id="req_quota_3",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="quota_exceeded",
            requested_at=now - timedelta(minutes=1),
        )
        await logs_repo.add_log(
            account_id="acc_quota",
            request_id="req_quota_4",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="rate_limit_exceeded",
            requested_at=now - timedelta(minutes=4),
        )

    response = await async_client.get("/api/request-logs?status=quota&limit=10")
    assert response.status_code == 200
    payload = response.json()["requests"]
    codes = {entry["errorCode"] for entry in payload}
    assert codes == {"insufficient_quota", "usage_not_included", "quota_exceeded"}
    assert all(entry["status"] == "quota" for entry in payload)


@pytest.mark.asyncio
async def test_request_logs_filters_by_account_model_and_time(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_filter", "filter@example.com"))

        await logs_repo.add_log(
            account_id="acc_filter",
            request_id="req_filter_1",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=20),
        )
        await logs_repo.add_log(
            account_id="acc_filter",
            request_id="req_filter_2",
            model="gpt-5.1",
            input_tokens=2,
            output_tokens=2,
            latency_ms=10,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=10),
        )
        await logs_repo.add_log(
            account_id="acc_filter",
            request_id="req_filter_3",
            model="gpt-5.2",
            input_tokens=3,
            output_tokens=3,
            latency_ms=10,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=5),
        )

    since = (now - timedelta(minutes=15)).isoformat()
    until = (now - timedelta(minutes=7)).isoformat()
    response = await async_client.get(
        f"/api/request-logs?accountId=acc_filter&model=gpt-5.1&since={since}&until={until}"
    )
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    assert payload[0]["model"] == "gpt-5.1"
    assert payload[0]["tokens"] == 4


@pytest.mark.asyncio
async def test_request_logs_expose_requested_and_actual_service_tiers(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_tier", "tiers@example.com"))

        await logs_repo.add_log(
            account_id="acc_tier",
            request_id="req_tier_1",
            model="gpt-5.4",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="success",
            error_code=None,
            service_tier="default",
            requested_service_tier="priority",
            actual_service_tier="default",
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert payload[0]["serviceTier"] == "default"
    assert payload[0]["requestedServiceTier"] == "priority"
    assert payload[0]["actualServiceTier"] == "default"


@pytest.mark.asyncio
async def test_request_logs_filters_by_multiple_accounts_returns_union(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_multi_a", "a@example.com"))
        await accounts_repo.upsert(_make_account("acc_multi_b", "b@example.com"))

        await logs_repo.add_log(
            account_id="acc_multi_a",
            request_id="req_multi_1",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=1),
        )
        await logs_repo.add_log(
            account_id="acc_multi_b",
            request_id="req_multi_2",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="success",
            error_code=None,
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs?accountId=acc_multi_a&accountId=acc_multi_b&limit=10")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert {entry["accountId"] for entry in payload} == {"acc_multi_a", "acc_multi_b"}


@pytest.mark.asyncio
async def test_request_logs_status_error_excludes_rate_limit_and_quota(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_err_only", "err-only@example.com"))

        await logs_repo.add_log(
            account_id="acc_err_only",
            request_id="req_err_rate_limit",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="rate_limit_exceeded",
            requested_at=now - timedelta(minutes=2),
        )
        await logs_repo.add_log(
            account_id="acc_err_only",
            request_id="req_err_quota",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="insufficient_quota",
            requested_at=now - timedelta(minutes=1),
        )
        await logs_repo.add_log(
            account_id="acc_err_only",
            request_id="req_err_other",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="upstream_error",
            error_message="upstream failure",
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs?status=error&limit=10")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    assert payload[0]["status"] == "error"
    assert payload[0]["errorCode"] == "upstream_error"


@pytest.mark.asyncio
async def test_request_logs_search_matches_email_and_error(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_search", "example@myemail.com"))

        await logs_repo.add_log(
            account_id="acc_search",
            request_id="req_search_1",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="success",
            error_code=None,
            requested_at=now - timedelta(minutes=1),
        )
        await logs_repo.add_log(
            account_id="acc_search",
            request_id="req_search_2",
            model="gpt-5.1",
            input_tokens=1,
            output_tokens=1,
            latency_ms=10,
            status="error",
            error_code="upstream_error",
            error_message="This is an example string",
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs?search=example&limit=50")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert {entry["requestId"] for entry in payload} == {"req_search_1", "req_search_2"}


@pytest.mark.asyncio
async def test_request_logs_tokens_and_cost_use_reasoning_tokens(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_reason", "reason@example.com"))

        await logs_repo.add_log(
            account_id="acc_reason",
            request_id="req_reason_1",
            model="gpt-5.1",
            input_tokens=1000,
            output_tokens=None,
            cached_input_tokens=100,
            reasoning_tokens=400,
            reasoning_effort="xhigh",
            latency_ms=50,
            status="success",
            error_code=None,
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs?accountId=acc_reason&limit=1")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    entry = payload[0]
    assert entry["tokens"] == 1400
    assert entry["cachedInputTokens"] == 100
    assert entry["reasoningEffort"] == "xhigh"
    expected = round(_cost(1000, 400, 100), 6)
    assert entry["costUsd"] == pytest.approx(expected)


async def test_request_logs_cost_uses_priority_service_tier(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_priority", "priority@example.com"))

        await logs_repo.add_log(
            account_id="acc_priority",
            request_id="req_priority_1",
            model="gpt-5.4",
            service_tier="priority",
            input_tokens=1_000_000,
            output_tokens=1_000_000,
            latency_ms=50,
            status="success",
            error_code=None,
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs?accountId=acc_priority&limit=1")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    entry = payload[0]
    assert entry["serviceTier"] == "priority"
    expected = round(_cost(1_000_000, 1_000_000, input_rate=5.0, cached_rate=0.5, output_rate=30.0), 6)
    assert entry["costUsd"] == pytest.approx(expected)


@pytest.mark.asyncio
async def test_request_logs_cost_uses_flex_service_tier(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_flex", "flex@example.com"))

        await logs_repo.add_log(
            account_id="acc_flex",
            request_id="req_flex_1",
            model="gpt-5.4",
            service_tier="flex",
            input_tokens=300_000,
            output_tokens=100_000,
            cached_input_tokens=50_000,
            latency_ms=50,
            status="success",
            error_code=None,
            requested_at=now,
        )

    response = await async_client.get("/api/request-logs?accountId=acc_flex&limit=1")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    entry = payload[0]
    assert entry["serviceTier"] == "flex"
    expected = round(_cost(300_000, 100_000, 50_000, input_rate=2.5, cached_rate=0.25, output_rate=11.25), 6)
    assert entry["costUsd"] == pytest.approx(expected)


@pytest.mark.asyncio
async def test_request_logs_cost_uses_persisted_cost_field(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_persisted_log_cost", "persisted-log-cost@example.com"))

        log = await logs_repo.add_log(
            account_id="acc_persisted_log_cost",
            request_id="req_persisted_log_cost_1",
            model="gpt-5.1",
            input_tokens=1000,
            output_tokens=500,
            latency_ms=50,
            status="success",
            error_code=None,
            requested_at=now,
        )
        await session.execute(update(log.__class__).where(log.__class__.id == log.id).values(cost_usd=4.321234))
        await session.commit()

    response = await async_client.get("/api/request-logs?accountId=acc_persisted_log_cost&limit=1")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    assert payload[0]["costUsd"] == pytest.approx(4.321234)


@pytest.mark.asyncio
async def test_request_logs_search_matches_api_key_name(async_client, db_setup):
    now = utcnow()
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)
        await accounts_repo.upsert(_make_account("acc_key_search", "key-search@example.com"))
        session.add(
            ApiKey(
                id="key_search_1",
                name="Window-Runner",
                key_hash="hash_key_search_1",
                key_prefix="sk-test",
            )
        )
        await session.commit()

        await logs_repo.add_log(
            account_id="acc_key_search",
            request_id="req_key_search_1",
            model="gpt-5.1",
            input_tokens=3,
            output_tokens=2,
            latency_ms=10,
            status="success",
            error_code=None,
            requested_at=now,
            api_key_id="key_search_1",
        )

    response = await async_client.get("/api/request-logs?search=window-runner&limit=50")
    assert response.status_code == 200
    payload = response.json()["requests"]
    assert len(payload) == 1
    assert payload[0]["requestId"] == "req_key_search_1"
    assert payload[0]["apiKeyName"] == "Window-Runner"



from __future__ import annotations

import time
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
import pytest

from app.core.clients.codex_version import CodexVersionCache


def _mock_response(*, status: int = 200, json_data: object = None) -> MagicMock:
    resp = MagicMock()
    resp.status = status
    resp.json = AsyncMock(return_value=json_data)
    resp.__aenter__ = AsyncMock(return_value=resp)
    resp.__aexit__ = AsyncMock(return_value=False)
    return resp


def _mock_session(response: MagicMock) -> MagicMock:
    session = MagicMock()
    session.get = MagicMock(return_value=response)
    session.__aenter__ = AsyncMock(return_value=session)
    session.__aexit__ = AsyncMock(return_value=False)
    return session


@pytest.mark.asyncio
async def test_fetches_version_from_github():
    cache = CodexVersionCache(ttl_seconds=60)
    resp = _mock_response(json_data={"name": "1.2.3", "tag_name": "rust-v1.2.3"})
    session = _mock_session(resp)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session) as client_session_cls:
        version = await cache.get_version()

    assert version == "1.2.3"
    assert client_session_cls.call_args.kwargs["trust_env"] is True


@pytest.mark.asyncio
async def test_returns_cached_version_within_ttl():
    cache = CodexVersionCache(ttl_seconds=60)
    resp = _mock_response(json_data={"name": "1.2.3"})
    session = _mock_session(resp)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session):
        first = await cache.get_version()
        second = await cache.get_version()

    assert first == second == "1.2.3"
    # Only one HTTP call — second hit served from cache
    session.get.assert_called_once()


@pytest.mark.asyncio
async def test_refetches_after_ttl_expires():
    cache = CodexVersionCache(ttl_seconds=60)
    resp1 = _mock_response(json_data={"name": "1.0.0"})
    session1 = _mock_session(resp1)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session1):
        v1 = await cache.get_version()
    assert v1 == "1.0.0"

    # Expire the cache
    cache._cached_at = time.monotonic() - 120

    resp2 = _mock_response(json_data={"name": "2.0.0"})
    session2 = _mock_session(resp2)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session2):
        v2 = await cache.get_version()
    assert v2 == "2.0.0"


@pytest.mark.asyncio
async def test_rejects_invalid_version_name():
    cache = CodexVersionCache(ttl_seconds=60)
    resp = _mock_response(json_data={"name": "rust-v1.2.3"})
    session = _mock_session(resp)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session):
        version = await cache.get_version()

    # Invalid name falls back to settings default
    assert version == "0.101.0"


@pytest.mark.asyncio
async def test_rejects_alpha_version_name():
    cache = CodexVersionCache(ttl_seconds=60)
    resp = _mock_response(json_data={"name": "1.2.3-alpha.1"})
    session = _mock_session(resp)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session):
        version = await cache.get_version()

    assert version == "0.101.0"


@pytest.mark.asyncio
async def test_fallback_to_stale_cache_on_github_failure():
    cache = CodexVersionCache(ttl_seconds=60)

    # Populate cache
    resp_ok = _mock_response(json_data={"name": "1.5.0"})
    session_ok = _mock_session(resp_ok)
    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session_ok):
        v1 = await cache.get_version()
    assert v1 == "1.5.0"

    # Expire the cache
    cache._cached_at = time.monotonic() - 120

    # GitHub returns error
    resp_fail = _mock_response(status=503, json_data=None)
    session_fail = _mock_session(resp_fail)
    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session_fail):
        v2 = await cache.get_version()

    # Returns stale cached value
    assert v2 == "1.5.0"


@pytest.mark.asyncio
async def test_fallback_to_settings_default_when_no_cache():
    cache = CodexVersionCache(ttl_seconds=60)

    resp_fail = _mock_response(status=500, json_data=None)
    session_fail = _mock_session(resp_fail)
    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session_fail):
        version = await cache.get_version()

    assert version == "0.101.0"


@pytest.mark.asyncio
async def test_fallback_on_network_exception():
    cache = CodexVersionCache(ttl_seconds=60)

    with patch(
        "app.core.clients.codex_version.aiohttp.ClientSession",
        side_effect=aiohttp.ClientError("connection refused"),
    ):
        version = await cache.get_version()

    assert version == "0.101.0"


@pytest.mark.asyncio
async def test_invalidate_clears_cache():
    cache = CodexVersionCache(ttl_seconds=60)
    resp = _mock_response(json_data={"name": "3.0.0"})
    session = _mock_session(resp)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session):
        v1 = await cache.get_version()
    assert v1 == "3.0.0"

    await cache.invalidate()

    resp2 = _mock_response(json_data={"name": "4.0.0"})
    session2 = _mock_session(resp2)
    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session2):
        v2 = await cache.get_version()
    assert v2 == "4.0.0"


@pytest.mark.asyncio
async def test_missing_name_field_falls_back():
    cache = CodexVersionCache(ttl_seconds=60)
    resp = _mock_response(json_data={"tag_name": "rust-v1.2.3"})
    session = _mock_session(resp)

    with patch("app.core.clients.codex_version.aiohttp.ClientSession", return_value=session):
        version = await cache.get_version()

    assert version == "0.101.0"


def test_ttl_must_be_positive():
    with pytest.raises(ValueError, match="ttl_seconds must be positive"):
        CodexVersionCache(ttl_seconds=0)
    with pytest.raises(ValueError, match="ttl_seconds must be positive"):
        CodexVersionCache(ttl_seconds=-1)



from __future__ import annotations

from pydantic import AliasChoices, BaseModel, ConfigDict, Field, StrictStr, field_validator

from app.core.types import JsonObject


class OAuthTokenPayload(BaseModel):
    model_config = ConfigDict(extra="ignore")

    access_token: StrictStr | None = None
    refresh_token: StrictStr | None = None
    id_token: StrictStr | None = None
    authorization_code: StrictStr | None = None
    code_verifier: StrictStr | None = None
    error: JsonObject | StrictStr | None = None
    error_description: StrictStr | None = None
    message: StrictStr | None = None
    error_code: StrictStr | None = None
    code: StrictStr | None = None
    status: StrictStr | None = None


class DeviceCodePayload(BaseModel):
    model_config = ConfigDict(extra="ignore")

    device_auth_id: StrictStr | None = None
    user_code: StrictStr | None = Field(
        default=None,
        validation_alias=AliasChoices("user_code", "usercode"),
    )
    interval: int | None = None
    expires_in: int | None = None
    expires_at: StrictStr | None = None

    @field_validator("interval", mode="before")
    @classmethod
    def _parse_interval(cls, value: object) -> int | None:
        if value is None:
            return None
        if isinstance(value, int):
            return value
        if isinstance(value, str):
            stripped = value.strip()
            if not stripped:
                return None
            if stripped.isdigit():
                return int(stripped)
        raise ValueError("Invalid interval")



#!/usr/bin/env python3
"""PostToolUse hook — track code file modifications.

Matcher: Edit|Write|MultiEdit

Records modified code files in dirty-{session_id}.json.
"""

import json
import sys
from datetime import UTC, datetime
from pathlib import Path

try:
    from _analytics import emit_event
except ImportError:

    def emit_event(*_a: object, **_k: object) -> None:
        pass


# Code file extensions (backend only)
CODE_EXTENSIONS = {".py"}

# Code directories (relative to project root)
CODE_DIRS = {"app/", "tests/"}

# Exclusion patterns
EXCLUDE_PATTERNS = {"__pycache__/", ".claude/", ".agents/"}


def is_code_file(file_path: str, project_dir: str) -> bool:
    """Determine if the given path is a code file."""
    if Path(file_path).suffix not in CODE_EXTENSIONS:
        return False

    try:
        rel = str(Path(file_path).relative_to(project_dir))
    except ValueError:
        return False

    if any(excl in rel for excl in EXCLUDE_PATTERNS):
        return False

    return any(rel.startswith(d) for d in CODE_DIRS)


def main() -> None:
    try:
        input_data = json.loads(sys.stdin.read())
    except (json.JSONDecodeError, EOFError):
        sys.exit(0)

    session_id = input_data.get("session_id", "unknown")
    tool_input = input_data.get("tool_input", {})
    file_path = tool_input.get("file_path", "")
    project_dir = input_data.get("cwd", "")

    if not file_path or not is_code_file(file_path, project_dir):
        sys.exit(0)

    # Record dirty state
    hook_dir = Path(__file__).resolve().parent
    state_dir = hook_dir / "state"
    state_dir.mkdir(parents=True, exist_ok=True)
    state_path = state_dir / f"dirty-{session_id}.json"

    # Load existing state
    state: dict[str, object] = {"modified": True, "files": [], "last_modified": ""}
    if state_path.exists():
        try:
            with open(state_path) as f:
                state = json.load(f)
        except (json.JSONDecodeError, PermissionError):
            pass

    # Add file (deduplicate)
    files = state.get("files", [])
    if isinstance(files, list) and file_path not in files:
        files.append(file_path)
    state["files"] = files
    state["modified"] = True
    state["last_modified"] = datetime.now(UTC).isoformat()

    with open(state_path, "w") as f:
        json.dump(state, f, indent=2)

    emit_event(
        session_id,
        "hook.invoked",
        {
            "hook": "post_edit_tracker",
            "trigger": "PostToolUse",
            "outcome": "tracked",
            "matched_count": 1,
            "exit_code": 0,
        },
    )
    sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import time
from asyncio import Semaphore

from starlette.types import ASGIApp, Receive, Scope, Send

from app.core.resilience.memory_monitor import is_memory_pressure, is_memory_warning

logger = logging.getLogger(__name__)

_MEMORY_WARNING_LOG_INTERVAL = 30.0
_last_memory_warning_log: float = 0.0


async def _reject_websocket(receive: Receive, send: Send, *, reason: str) -> None:
    try:
        event = await receive()
        if event.get("type") != "websocket.connect":
            return
    except Exception:
        return
    await send({"type": "websocket.close", "code": 1013, "reason": reason})


class BulkheadSemaphore:
    def __init__(self, proxy_limit: int = 200, dashboard_limit: int = 50, background_limit: int = 10) -> None:
        self._proxy = Semaphore(proxy_limit) if proxy_limit > 0 else None
        self._dashboard = Semaphore(dashboard_limit) if dashboard_limit > 0 else None
        self._background = Semaphore(background_limit) if background_limit > 0 else None

    def get_semaphore(self, path: str) -> Semaphore | None:
        if path.startswith("/v1/") or path.startswith("/backend-api/"):
            return self._proxy
        return self._dashboard


class BulkheadMiddleware:
    def __init__(self, app: ASGIApp, *, bulkhead: BulkheadSemaphore) -> None:
        self.app = app
        self._bulkhead = bulkhead

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        path = scope.get("path", "")

        if path.startswith("/health"):
            await self.app(scope, receive, send)
            return

        if is_memory_warning():
            global _last_memory_warning_log
            now = time.monotonic()
            if now - _last_memory_warning_log >= _MEMORY_WARNING_LOG_INTERVAL:
                _last_memory_warning_log = now
                logger.warning("Memory warning threshold exceeded")

        if is_memory_pressure():
            if scope["type"] == "websocket":
                await _reject_websocket(receive, send, reason="Service temporarily unavailable (memory pressure)")
                return
            body = b'{"detail":"Service temporarily unavailable (memory pressure)"}'
            await send(
                {
                    "type": "http.response.start",
                    "status": 503,
                    "headers": [(b"retry-after", b"5"), (b"content-type", b"application/json")],
                }
            )
            await send({"type": "http.response.body", "body": body})
            return

        sem = self._bulkhead.get_semaphore(path)
        if sem is None:
            await self.app(scope, receive, send)
            return

        if sem._value <= 0:
            if scope["type"] == "websocket":
                await _reject_websocket(receive, send, reason="Service temporarily unavailable (bulkhead full)")
                return
            body = b'{"detail":"Service temporarily unavailable (bulkhead full)"}'
            await send(
                {
                    "type": "http.response.start",
                    "status": 503,
                    "headers": [(b"retry-after", b"5"), (b"content-type", b"application/json")],
                }
            )
            await send({"type": "http.response.body", "body": body})
            return

        await sem.acquire()

        try:
            await self.app(scope, receive, send)
        finally:
            sem.release()


_bulkhead: BulkheadSemaphore | None = None


def get_bulkhead(proxy_limit: int = 200, dashboard_limit: int = 50) -> BulkheadSemaphore:
    global _bulkhead
    if _bulkhead is None:
        _bulkhead = BulkheadSemaphore(proxy_limit=proxy_limit, dashboard_limit=dashboard_limit)
    return _bulkhead


__all__ = ["BulkheadMiddleware", "BulkheadSemaphore", "get_bulkhead"]



from __future__ import annotations

import base64
import json

import pytest

import app.modules.proxy.service as proxy_module

pytestmark = pytest.mark.integration


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str, email: str) -> dict:
    payload = {
        "email": email,
        "chatgpt_account_id": account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    return {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access-token",
            "refreshToken": "refresh-token",
            "accountId": account_id,
        },
    }


async def _import_account(async_client, account_id: str, email: str) -> None:
    auth_json = _make_auth_json(account_id, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200


def _completed_event(response_id: str) -> str:
    return 'data: {"type":"response.completed","response":{"id":"' + response_id + '","status":"completed"}}\n\n'


@pytest.mark.asyncio
async def test_v1_responses_forwards_input_file_url(async_client, monkeypatch):
    await _import_account(async_client, "acc_file_url", "file-url@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_file_url")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "input": [
            {
                "role": "user",
                "content": [
                    {"type": "input_text", "text": "Summarize this file."},
                    {"type": "input_file", "file_url": "https://example.com/file.pdf"},
                ],
            }
        ],
    }
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 200
    assert seen["payload"].input == payload["input"]


@pytest.mark.asyncio
async def test_v1_responses_rejects_input_file_id(async_client):
    payload = {
        "model": "gpt-5.2",
        "input": [
            {
                "role": "user",
                "content": [
                    {"type": "input_text", "text": "Summarize this file."},
                    {"type": "input_file", "file_id": "file-123"},
                ],
            }
        ],
    }
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 400
    payload = resp.json()
    assert payload["error"]["type"] == "invalid_request_error"
    assert payload["error"]["message"] == "Invalid request payload"
    assert payload["error"]["param"] == "input"


@pytest.mark.asyncio
async def test_v1_responses_accepts_previous_response_id(async_client, monkeypatch):
    await _import_account(async_client, "acc_prev_response_id", "prev-response-id@example.com")
    seen_previous_response_ids: list[str | None] = []

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False, **_kw):
        del headers, access_token, account_id, base_url, raise_for_status, _kw
        seen_previous_response_ids.append(getattr(payload, "previous_response_id", None))
        yield 'data: {"type":"response.completed","response":{"id":"resp_abc123"}}\n\n'

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "previous_response_id": "resp_abc123",
        "input": [
            {
                "role": "user",
                "content": [{"type": "input_text", "text": "Continue."}],
            }
        ],
        "stream": True,
    }
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 200
    assert seen_previous_response_ids == ["resp_abc123"]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tool_payload",
    [
        {"type": "file_search", "vector_store_ids": ["vs_dummy"]},
        {"type": "code_interpreter", "container": {"type": "auto"}},
        {
            "type": "computer_use_preview",
            "display_width": 1024,
            "display_height": 768,
            "environment": "browser",
        },
        {"type": "image_generation"},
    ],
)
async def test_v1_responses_rejects_builtin_tools(async_client, tool_payload):
    request_payload = {
        "model": "gpt-5.2",
        "input": [
            {
                "role": "user",
                "content": [{"type": "input_text", "text": "Run tool."}],
            }
        ],
        "tools": [tool_payload],
    }

    resp = await async_client.post("/v1/responses", json=request_payload)
    assert resp.status_code == 400
    assert resp.json()["error"]["type"] == "invalid_request_error"


@pytest.mark.asyncio
async def test_v1_responses_forwards_input_string(async_client, monkeypatch):
    await _import_account(async_client, "acc_input_string", "input-string@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_input_string")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {"model": "gpt-5.2", "input": "Hello"}
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 200
    assert seen["payload"].input == [
        {"role": "user", "content": [{"type": "input_text", "text": "Hello"}]},
    ]


@pytest.mark.asyncio
async def test_v1_responses_forwards_include_logprobs(async_client, monkeypatch):
    await _import_account(async_client, "acc_include_logprobs", "include-logprobs@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_include")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "input": [{"role": "user", "content": [{"type": "input_text", "text": "hi"}]}],
        "include": ["message.output_text.logprobs"],
    }
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 200
    assert seen["payload"].include == ["message.output_text.logprobs"]


@pytest.mark.asyncio
async def test_v1_responses_preserves_prompt_cache_controls(async_client, monkeypatch):
    await _import_account(async_client, "acc_prompt_cache_v1", "prompt-cache-v1@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload.to_payload()
        yield _completed_event("resp_prompt_cache_v1")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "input": "cache me",
        "prompt_cache_key": "thread_123",
        "prompt_cache_retention": "4h",
    }
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 200
    assert seen["payload"]["prompt_cache_key"] == "thread_123"
    assert "prompt_cache_retention" not in seen["payload"]


@pytest.mark.asyncio
async def test_v1_responses_normalizes_prompt_cache_aliases(async_client, monkeypatch):
    await _import_account(async_client, "acc_prompt_cache_alias", "prompt-cache-alias@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload.to_payload()
        yield _completed_event("resp_prompt_cache_alias")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "input": "cache me",
        "promptCacheKey": "thread_alias",
        "promptCacheRetention": "12h",
    }
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 200
    assert seen["payload"]["prompt_cache_key"] == "thread_alias"
    assert "prompt_cache_retention" not in seen["payload"]
    assert "promptCacheKey" not in seen["payload"]
    assert "promptCacheRetention" not in seen["payload"]


@pytest.mark.asyncio
async def test_backend_responses_forwards_service_tier(async_client, monkeypatch):
    await _import_account(async_client, "acc_backend_service_tier", "backend-service-tier@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_backend_service_tier")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    request_payload = {
        "model": "gpt-5.2",
        "instructions": "",
        "input": [{"role": "user", "content": [{"type": "input_text", "text": "Fast"}]}],
        "service_tier": "priority",
    }
    resp = await async_client.post("/backend-api/codex/responses", json=request_payload)
    assert resp.status_code == 200
    assert seen["payload"].service_tier == "priority"


@pytest.mark.asyncio
async def test_backend_responses_normalizes_fast_service_tier_for_upstream(async_client, monkeypatch):
    await _import_account(async_client, "acc_backend_fast_tier", "backend-fast-tier@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload.to_payload()
        yield _completed_event("resp_backend_fast_tier")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    request_payload = {
        "model": "gpt-5.2",
        "instructions": "",
        "input": [{"role": "user", "content": [{"type": "input_text", "text": "Fast"}]}],
        "service_tier": "fast",
    }
    resp = await async_client.post("/backend-api/codex/responses", json=request_payload)
    assert resp.status_code == 200
    assert seen["payload"]["service_tier"] == "priority"


@pytest.mark.asyncio
async def test_v1_responses_rejects_invalid_include(async_client):
    payload = {"model": "gpt-5.2", "input": "hi", "include": ["not_allowed"]}
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
async def test_v1_responses_rejects_store_true(async_client):
    payload = {"model": "gpt-5.2", "input": "hi", "store": True}
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
@pytest.mark.parametrize("truncation", ["auto", "disabled"])
async def test_v1_responses_rejects_truncation(async_client, truncation):
    payload = {"model": "gpt-5.2", "input": "hi", "truncation": truncation}
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
async def test_v1_responses_rejects_conversation_and_previous(async_client):
    payload = {
        "model": "gpt-5.2",
        "input": "hi",
        "conversation": "conv_1",
        "previous_response_id": "resp_1",
    }
    resp = await async_client.post("/v1/responses", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
@pytest.mark.parametrize("tool_type", ["web_search", "web_search_preview"])
async def test_v1_responses_allows_web_search(async_client, monkeypatch, tool_type):
    await _import_account(async_client, "acc_web_search", "web-search@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_web_search")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    request_payload = {
        "model": "gpt-5.2",
        "input": [{"role": "user", "content": [{"type": "input_text", "text": "Search"}]}],
        "tools": [{"type": tool_type}],
    }
    resp = await async_client.post("/v1/responses", json=request_payload)
    assert resp.status_code == 200
    assert seen["payload"].tools == [{"type": "web_search"}]


@pytest.mark.asyncio
@pytest.mark.parametrize("tool_type", ["web_search", "web_search_preview"])
async def test_backend_responses_allows_web_search(async_client, monkeypatch, tool_type):
    await _import_account(async_client, "acc_backend_web_search", "backend-web-search@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_backend_web_search")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    request_payload = {
        "model": "gpt-5.2",
        "instructions": "",
        "input": [{"role": "user", "content": [{"type": "input_text", "text": "Search"}]}],
        "tools": [{"type": tool_type}],
    }
    resp = await async_client.post("/backend-api/codex/responses", json=request_payload)
    assert resp.status_code == 200
    assert seen["payload"].tools == [{"type": "web_search"}]


@pytest.mark.asyncio
async def test_v1_chat_completions_rejects_non_text_developer(async_client):
    payload = {
        "model": "gpt-5.2",
        "messages": [
            {
                "role": "developer",
                "content": [{"type": "image_url", "image_url": {"url": "https://example.com/a.png"}}],
            },
            {"role": "user", "content": "hi"},
        ],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
async def test_v1_chat_completions_rejects_invalid_audio(async_client):
    payload = {
        "model": "gpt-5.2",
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "input_audio", "input_audio": {"data": "AAA", "format": "ogg"}},
                ],
            }
        ],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
async def test_v1_chat_completions_maps_response_format(async_client, monkeypatch):
    await _import_account(async_client, "acc_chat_format", "chat-format@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_chat_format")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Return JSON."}],
        "response_format": {
            "type": "json_schema",
            "json_schema": {
                "name": "result_schema",
                "schema": {"type": "object", "properties": {"ok": {"type": "boolean"}}},
                "strict": True,
            },
        },
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 200
    text = seen["payload"].text
    assert text is not None
    assert text.format is not None
    assert text.format.type == "json_schema"
    assert text.format.name == "result_schema"


@pytest.mark.asyncio
async def test_v1_chat_completions_rejects_missing_json_schema(async_client):
    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Return JSON."}],
        "response_format": {"type": "json_schema"},
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
async def test_v1_chat_completions_forwards_multimodal(async_client, monkeypatch):
    await _import_account(async_client, "acc_chat_multi", "chat-multi@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_chat_multi")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Check image and audio."},
                    {"type": "image_url", "image_url": {"url": "https://example.com/a.png"}},
                    {"type": "file", "file": {"file_url": "https://example.com/file.pdf"}},
                ],
            }
        ],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 200
    assert isinstance(seen["payload"].input, list)
    assert seen["payload"].input[0]["role"] == "user"
    content = seen["payload"].input[0]["content"]
    assert content[0] == {"type": "input_text", "text": "Check image and audio."}
    assert content[1] == {"type": "input_image", "image_url": "https://example.com/a.png"}
    assert content[2] == {"type": "input_file", "file_url": "https://example.com/file.pdf"}


@pytest.mark.asyncio
async def test_v1_chat_completions_rejects_file_id(async_client):
    payload = {
        "model": "gpt-5.2",
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Summarize file."},
                    {"type": "file", "file": {"file_id": "file-123"}},
                ],
            }
        ],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 400
    payload = resp.json()
    assert payload["error"]["type"] == "invalid_request_error"
    assert payload["error"]["message"] == "Invalid request payload"
    assert payload["error"]["param"] == "messages"


@pytest.mark.asyncio
async def test_v1_chat_completions_rejects_audio_input(async_client):
    payload = {
        "model": "gpt-5.2",
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Transcribe audio."},
                    {"type": "input_audio", "input_audio": {"data": "AAA", "format": "wav"}},
                ],
            }
        ],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
async def test_v1_chat_completions_rejects_builtin_tools(async_client):
    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Search the web."}],
        "tools": [{"type": "image_generation"}],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 400


@pytest.mark.asyncio
@pytest.mark.parametrize("tool_type", ["web_search", "web_search_preview"])
async def test_v1_chat_completions_allows_web_search(async_client, monkeypatch, tool_type):
    await _import_account(async_client, "acc_chat_web_search", "chat-web-search@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_chat_web_search")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Search the web."}],
        "tools": [{"type": tool_type}],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 200
    assert seen["payload"].tools == [{"type": "web_search"}]


@pytest.mark.asyncio
async def test_v1_chat_completions_normalizes_tools_and_tool_choice(async_client, monkeypatch):
    await _import_account(async_client, "acc_chat_tools", "chat-tools@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_chat_tools")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Weather?"}],
        "tools": [
            {
                "type": "function",
                "function": {
                    "name": "get_weather",
                    "description": "Get weather",
                    "parameters": {"type": "object", "properties": {"city": {"type": "string"}}},
                },
            }
        ],
        "tool_choice": {"type": "function", "function": {"name": "get_weather"}},
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 200
    assert seen["payload"].tools == [
        {
            "type": "function",
            "name": "get_weather",
            "description": "Get weather",
            "parameters": {"type": "object", "properties": {"city": {"type": "string"}}},
        }
    ]
    assert seen["payload"].tool_choice == {"type": "function", "name": "get_weather"}


@pytest.mark.asyncio
async def test_v1_chat_completions_does_not_enable_codex_session_affinity(async_client, monkeypatch):
    await _import_account(async_client, "acc_chat_affinity_a", "chat-affinity-a@example.com")
    await _import_account(async_client, "acc_chat_affinity_b", "chat-affinity-b@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["account_id"] = account_id
        seen["prompt_cache_key"] = getattr(payload, "prompt_cache_key", None)
        yield _completed_event("resp_chat_affinity")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Weather?"}],
    }
    resp = await async_client.post("/v1/chat/completions", json=payload, headers={"session_id": "chat-session-123"})
    assert resp.status_code == 200
    assert isinstance(seen["prompt_cache_key"], str)
    assert seen["prompt_cache_key"]


@pytest.mark.asyncio
async def test_v1_chat_completions_maps_reasoning_effort(async_client, monkeypatch):
    await _import_account(async_client, "acc_chat_reason", "chat-reason@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_chat_reason")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Think."}],
        "reasoning_effort": "low",
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 200
    assert seen["payload"].reasoning is not None
    assert seen["payload"].reasoning.effort == "low"


@pytest.mark.asyncio
async def test_v1_chat_completions_forwards_service_tier(async_client, monkeypatch):
    await _import_account(async_client, "acc_chat_service_tier", "chat-service-tier@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload
        yield _completed_event("resp_chat_service_tier")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Think fast."}],
        "service_tier": "priority",
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 200
    assert seen["payload"].service_tier == "priority"


@pytest.mark.asyncio
async def test_v1_chat_completions_preserves_prompt_cache_controls(async_client, monkeypatch):
    await _import_account(async_client, "acc_chat_prompt_cache", "chat-prompt-cache@example.com")

    seen = {}

    async def fake_stream(payload, headers, access_token, account_id, base_url=None, raise_for_status=False):
        seen["payload"] = payload.to_payload()
        yield _completed_event("resp_chat_prompt_cache")

    monkeypatch.setattr(proxy_module, "core_stream_responses", fake_stream)

    payload = {
        "model": "gpt-5.2",
        "messages": [{"role": "user", "content": "Cache this chat."}],
        "prompt_cache_key": "chat_thread_123",
        "prompt_cache_retention": "8h",
    }
    resp = await async_client.post("/v1/chat/completions", json=payload)
    assert resp.status_code == 200
    assert seen["payload"]["prompt_cache_key"] == "chat_thread_123"
    assert "prompt_cache_retention" not in seen["payload"]



from __future__ import annotations

import contextvars
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta

import aiohttp
from pydantic import ValidationError

from app.core.auth import OpenAIAuthClaims, extract_id_token_claims
from app.core.auth.models import OAuthTokenPayload
from app.core.balancer import PERMANENT_FAILURE_CODES
from app.core.clients.http import get_http_client
from app.core.config.settings import get_settings
from app.core.types import JsonObject
from app.core.utils.request_id import get_request_id
from app.core.utils.time import to_utc_naive, utcnow

TOKEN_REFRESH_INTERVAL_DAYS = 8

logger = logging.getLogger(__name__)
_TOKEN_REFRESH_TIMEOUT_OVERRIDE: contextvars.ContextVar[float | None] = contextvars.ContextVar(
    "token_refresh_timeout_override",
    default=None,
)


@dataclass(frozen=True)
class TokenRefreshResult:
    access_token: str
    refresh_token: str
    id_token: str
    account_id: str | None
    plan_type: str | None
    email: str | None


class RefreshError(Exception):
    def __init__(self, code: str, message: str, is_permanent: bool) -> None:
        super().__init__(message)
        self.code = code
        self.message = message
        self.is_permanent = is_permanent


def should_refresh(last_refresh: datetime, now: datetime | None = None) -> bool:
    current = to_utc_naive(now) if now is not None else utcnow()
    last = to_utc_naive(last_refresh)
    interval_days = get_settings().token_refresh_interval_days or TOKEN_REFRESH_INTERVAL_DAYS
    return current - last > timedelta(days=interval_days)


def classify_refresh_error(code: str | None) -> bool:
    if not code:
        return False
    return code in PERMANENT_FAILURE_CODES


async def refresh_access_token(
    refresh_token: str,
    *,
    session: aiohttp.ClientSession | None = None,
) -> TokenRefreshResult:
    settings = get_settings()
    url = f"{settings.auth_base_url.rstrip('/')}/oauth/token"
    payload = {
        "grant_type": "refresh_token",
        "client_id": settings.oauth_client_id,
        "refresh_token": refresh_token,
        "scope": settings.oauth_scope,
    }
    timeout = aiohttp.ClientTimeout(total=_effective_token_refresh_timeout(settings.token_refresh_timeout_seconds))

    client_session = session or get_http_client().session
    headers: dict[str, str] = {}
    request_id = get_request_id()
    if request_id:
        headers["x-request-id"] = request_id
    async with client_session.post(url, json=payload, headers=headers, timeout=timeout) as resp:
        data = await _safe_json(resp)
        try:
            payload_data = OAuthTokenPayload.model_validate(data)
        except ValidationError as exc:
            logger.warning(
                "Token refresh response invalid request_id=%s",
                get_request_id(),
            )
            raise RefreshError("invalid_response", "Refresh response invalid", False) from exc
        if resp.status >= 400:
            logger.warning(
                "Token refresh failed request_id=%s status=%s",
                get_request_id(),
                resp.status,
            )
            raise _refresh_error_from_payload(payload_data, resp.status)

    if not payload_data.access_token or not payload_data.refresh_token or not payload_data.id_token:
        raise RefreshError("invalid_response", "Refresh response missing tokens", False)

    claims = extract_id_token_claims(payload_data.id_token)
    auth_claims = claims.auth or OpenAIAuthClaims()
    account_id = auth_claims.chatgpt_account_id or claims.chatgpt_account_id
    plan_type = auth_claims.chatgpt_plan_type or claims.chatgpt_plan_type
    email = claims.email

    return TokenRefreshResult(
        access_token=payload_data.access_token,
        refresh_token=payload_data.refresh_token,
        id_token=payload_data.id_token,
        account_id=account_id,
        plan_type=plan_type,
        email=email,
    )


def push_token_refresh_timeout_override(timeout_seconds: float | None) -> contextvars.Token[float | None]:
    return _TOKEN_REFRESH_TIMEOUT_OVERRIDE.set(timeout_seconds)


def pop_token_refresh_timeout_override(token: contextvars.Token[float | None]) -> None:
    _TOKEN_REFRESH_TIMEOUT_OVERRIDE.reset(token)


async def _safe_json(resp: aiohttp.ClientResponse) -> JsonObject:
    try:
        data = await resp.json(content_type=None)
    except Exception:
        text = await resp.text()
        return {"error": {"message": text.strip()}}
    return data if isinstance(data, dict) else {"error": {"message": str(data)}}


def _refresh_error_from_payload(payload: OAuthTokenPayload, status_code: int) -> RefreshError:
    code = _extract_error_code(payload) or f"http_{status_code}"
    message = _extract_error_message(payload) or f"Token refresh failed ({status_code})"
    return RefreshError(code, message, classify_refresh_error(code))


def _effective_token_refresh_timeout(configured_timeout_seconds: float) -> float:
    override = _TOKEN_REFRESH_TIMEOUT_OVERRIDE.get()
    if override is None:
        return configured_timeout_seconds
    return max(0.001, min(configured_timeout_seconds, override))


def _extract_error_code(payload: OAuthTokenPayload) -> str | None:
    error = payload.error
    if isinstance(error, dict):
        code = error.get("code") or error.get("error")
        return code if isinstance(code, str) else None
    if isinstance(error, str):
        return error
    return payload.error_code or payload.code


def _extract_error_message(payload: OAuthTokenPayload) -> str | None:
    error = payload.error
    if isinstance(error, dict):
        message = error.get("message") or error.get("error_description")
        return message if isinstance(message, str) else None
    if isinstance(error, str):
        return payload.error_description or error
    return payload.message



from __future__ import annotations

import base64
import inspect
import json
from collections.abc import Mapping, Sequence

import pytest
import pytest_asyncio
from httpx import ASGITransport, AsyncClient

from app.core.openai.model_registry import ReasoningLevel, UpstreamModel, get_model_registry
from app.main import create_app

DEFAULT_DASHBOARD_PASSWORD = "password123"
DEFAULT_MODEL = "gpt-5.2"


def _encode_jwt(payload: Mapping[str, object]) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str, email: str) -> dict[str, object]:
    payload = {
        "email": email,
        "chatgpt_account_id": account_id,
        "https://api.openai.com/auth": {"chatgpt_plan_type": "plus"},
    }
    return {
        "tokens": {
            "idToken": _encode_jwt(payload),
            "accessToken": "access-token",
            "refreshToken": "refresh-token",
            "accountId": account_id,
        },
    }


def _make_upstream_model(slug: str) -> UpstreamModel:
    return UpstreamModel(
        slug=slug,
        display_name=slug,
        description=f"Test model {slug}",
        context_window=272000,
        input_modalities=("text",),
        supported_reasoning_levels=(ReasoningLevel(effort="medium", description="default"),),
        default_reasoning_level="medium",
        supports_reasoning_summaries=False,
        support_verbosity=False,
        default_verbosity=None,
        prefer_websockets=False,
        supports_parallel_tool_calls=True,
        supported_in_api=True,
        minimal_client_version=None,
        priority=0,
        available_in_plans=frozenset({"plus", "pro"}),
        raw={},
    )


async def _maybe_await(result: object) -> None:
    if inspect.isawaitable(result):
        await result


@pytest_asyncio.fixture
async def e2e_client(db_setup):
    app = create_app()
    async with app.router.lifespan_context(app):
        transport = ASGITransport(app=app)
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            yield client


@pytest_asyncio.fixture
async def client(e2e_client):
    yield e2e_client


@pytest.fixture
def dashboard_password() -> str:
    return DEFAULT_DASHBOARD_PASSWORD


@pytest.fixture
def setup_dashboard_password(dashboard_password: str):
    async def _setup(client: AsyncClient, *, password: str = dashboard_password) -> str:
        response = await client.post(
            "/api/dashboard-auth/password/setup",
            json={"password": password},
        )
        assert response.status_code == 200
        assert response.json()["passwordRequired"] is True
        return password

    return _setup


@pytest.fixture
def login_dashboard():
    async def _login(client: AsyncClient, *, password: str) -> dict[str, object]:
        response = await client.post(
            "/api/dashboard-auth/password/login",
            json={"password": password},
        )
        assert response.status_code == 200
        payload = response.json()
        assert payload["authenticated"] is True
        return payload

    return _login


@pytest.fixture
def enable_api_key_auth():
    async def _enable(client: AsyncClient) -> dict[str, object]:
        response = await client.put(
            "/api/settings",
            json={
                "stickyThreadsEnabled": False,
                "preferEarlierResetAccounts": False,
                "totpRequiredOnLogin": False,
                "apiKeyAuthEnabled": True,
            },
        )
        assert response.status_code == 200
        payload = response.json()
        assert payload["apiKeyAuthEnabled"] is True
        return payload

    return _enable


@pytest.fixture
def create_api_key():
    async def _create(
        client: AsyncClient,
        *,
        name: str,
        allowed_models: Sequence[str] | None = None,
        limits: list[dict[str, object]] | None = None,
    ) -> dict[str, object]:
        payload: dict[str, object] = {"name": name}
        if allowed_models is not None:
            payload["allowedModels"] = list(allowed_models)
        if limits is not None:
            payload["limits"] = limits

        response = await client.post("/api/api-keys/", json=payload)
        assert response.status_code == 200
        created = response.json()
        assert created["key"].startswith("sk-clb-")
        return created

    return _create


@pytest.fixture
def import_test_account():
    async def _import(client: AsyncClient, *, account_id: str, email: str) -> None:
        files = {
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json(account_id, email)),
                "application/json",
            )
        }
        response = await client.post("/api/accounts/import", files=files)
        assert response.status_code == 200

    return _import


@pytest.fixture
def populate_test_registry():
    async def _populate(models: Sequence[str] | None = None) -> list[str]:
        model_ids = list(models or [DEFAULT_MODEL])
        registry = get_model_registry()
        snapshot = {
            "plus": [_make_upstream_model(slug) for slug in model_ids],
            "pro": [_make_upstream_model(slug) for slug in model_ids],
        }
        await _maybe_await(registry.update(snapshot))
        return model_ids

    return _populate



from __future__ import annotations

from datetime import datetime

from pydantic import Field

from app.modules.shared.schemas import DashboardModel


class RequestLogEntry(DashboardModel):
    requested_at: datetime
    account_id: str | None = None
    provider_kind: str | None = None
    routing_subject_id: str | None = None
    api_key_name: str | None = None
    request_id: str
    model: str
    transport: str | None = None
    route_class: str | None = None
    upstream_request_id: str | None = None
    rejection_reason: str | None = None
    service_tier: str | None = None
    requested_service_tier: str | None = None
    actual_service_tier: str | None = None
    status: str
    error_code: str | None = None
    error_message: str | None = None
    tokens: int | None = None
    cached_input_tokens: int | None = None
    reasoning_effort: str | None = None
    cost_usd: float | None = None
    latency_ms: int | None = None
    latency_first_token_ms: int | None = None


class RequestLogsResponse(DashboardModel):
    requests: list[RequestLogEntry] = Field(default_factory=list)
    total: int
    has_more: bool


class RequestLogModelOption(DashboardModel):
    model: str
    reasoning_effort: str | None = None


class RequestLogFilterOptionsResponse(DashboardModel):
    account_ids: list[str] = Field(default_factory=list)
    model_options: list[RequestLogModelOption] = Field(default_factory=list)
    statuses: list[str] = Field(default_factory=list)



from __future__ import annotations

from app.core.usage.types import UsageTrendBucket
from app.modules.accounts.mappers import build_account_usage_trends


def _bucket(epoch: int, account_id: str, window: str, avg_used: float, samples: int = 1) -> UsageTrendBucket:
    return UsageTrendBucket(
        bucket_epoch=epoch,
        account_id=account_id,
        window=window,
        avg_used_percent=avg_used,
        samples=samples,
    )


# Use a value already aligned to BUCKET_SECONDS so tests are predictable
BUCKET_SECONDS = 21600  # 6h
SINCE_EPOCH = (1_700_000_000 // BUCKET_SECONDS) * BUCKET_SECONDS
BUCKET_COUNT = 4  # 4 buckets → spans 24h


class TestBuildAccountUsageTrends:
    def test_empty_buckets_returns_empty(self):
        result = build_account_usage_trends([], SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)
        assert result == {}

    def test_single_account_single_window(self):
        buckets = [
            _bucket(SINCE_EPOCH, "a1", "primary", 20.0),
            _bucket(SINCE_EPOCH + BUCKET_SECONDS, "a1", "primary", 40.0),
        ]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)

        assert "a1" in result
        trend = result["a1"]
        assert len(trend.primary) == BUCKET_COUNT

        # First bucket: 100 - 20 = 80
        assert trend.primary[0].v == 80.0
        # Second bucket: 100 - 40 = 60
        assert trend.primary[1].v == 60.0
        # Third and fourth buckets: forward-filled with last value (60)
        assert trend.primary[2].v == 60.0
        assert trend.primary[3].v == 60.0

    def test_values_are_remaining_percent(self):
        buckets = [_bucket(SINCE_EPOCH, "a1", "primary", 75.0)]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)
        assert result["a1"].primary[0].v == 25.0

    def test_used_percent_clamped_to_0_100(self):
        buckets = [_bucket(SINCE_EPOCH, "a1", "primary", 110.0)]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)
        # 100 - 110 = -10, clamped to 0
        assert result["a1"].primary[0].v == 0.0

    def test_missing_buckets_filled_with_default(self):
        # No data at all for bucket 0, data at bucket 1
        buckets = [_bucket(SINCE_EPOCH + BUCKET_SECONDS, "a1", "primary", 50.0)]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)

        # Bucket 0: no data → default 100.0
        assert result["a1"].primary[0].v == 100.0
        # Bucket 1: 100 - 50 = 50
        assert result["a1"].primary[1].v == 50.0

    def test_dual_window(self):
        buckets = [
            _bucket(SINCE_EPOCH, "a1", "primary", 20.0),
            _bucket(SINCE_EPOCH, "a1", "secondary", 30.0),
        ]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)
        trend = result["a1"]
        assert trend.primary[0].v == 80.0
        assert trend.secondary[0].v == 70.0

    def test_multiple_accounts(self):
        buckets = [
            _bucket(SINCE_EPOCH, "a1", "primary", 10.0),
            _bucket(SINCE_EPOCH, "a2", "primary", 90.0),
        ]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)
        assert result["a1"].primary[0].v == 90.0
        assert result["a2"].primary[0].v == 10.0

    def test_missing_window_returns_empty_list(self):
        buckets = [_bucket(SINCE_EPOCH, "a1", "primary", 20.0)]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)
        # secondary was not in any bucket → empty list
        assert result["a1"].secondary == []

    def test_timestamps_are_utc(self):
        buckets = [_bucket(SINCE_EPOCH, "a1", "primary", 0.0)]
        result = build_account_usage_trends(buckets, SINCE_EPOCH, BUCKET_SECONDS, BUCKET_COUNT)
        for point in result["a1"].primary:
            assert point.t.tzinfo is not None



from __future__ import annotations

from ipaddress import ip_network
from types import SimpleNamespace

import pytest
from starlette.requests import Request

import app.core.request_locality as request_locality
from app.core.request_locality import is_host_os_request, is_local_request


def _request(*, client_host: str, host: str) -> Request:
    scope = {
        "type": "http",
        "method": "GET",
        "path": "/",
        "headers": [(b"host", host.encode("utf-8"))],
        "client": (client_host, 50000),
        "server": (host.split(":", 1)[0], 80),
        "scheme": "http",
        "query_string": b"",
    }
    return Request(scope)


def test_loopback_with_local_host_is_local() -> None:
    request = _request(client_host="127.0.0.1", host="localhost")
    assert is_local_request(request) is True


def test_loopback_with_non_local_host_is_not_local() -> None:
    request = _request(client_host="127.0.0.1", host="lb.example")
    assert is_local_request(request) is False


def test_loopback_with_bracketed_ipv6_local_host_is_local() -> None:
    request = _request(client_host="::1", host="[::1]:8000")
    assert is_local_request(request) is True


def test_loopback_with_unbracketed_ipv6_local_host_is_local() -> None:
    request = _request(client_host="::1", host="::1")
    assert is_local_request(request) is True


def test_trusted_proxy_mode_treats_loopback_without_forwarded_hint_as_remote(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        request_locality,
        "get_settings",
        lambda: SimpleNamespace(firewall_trust_proxy_headers=True, firewall_trusted_proxy_cidrs=[]),
    )
    request = _request(client_host="127.0.0.1", host="localhost")
    assert is_local_request(request) is False


def test_trusted_proxy_mode_accepts_loopback_with_forwarded_hint(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        request_locality,
        "get_settings",
        lambda: SimpleNamespace(firewall_trust_proxy_headers=True, firewall_trusted_proxy_cidrs=[]),
    )
    scope = {
        "type": "http",
        "method": "GET",
        "path": "/",
        "headers": [(b"host", b"localhost"), (b"x-forwarded-for", b"127.0.0.1")],
        "client": ("127.0.0.1", 50000),
        "server": ("localhost", 80),
        "scheme": "http",
        "query_string": b"",
    }
    request = Request(scope)
    assert is_local_request(request) is True


def test_host_os_request_accepts_explicit_host_gateway_cidr(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        request_locality,
        "get_settings",
        lambda: SimpleNamespace(
            firewall_trust_proxy_headers=False,
            firewall_trusted_proxy_cidrs=[],
            insecure_allow_remote_no_auth_host_cidrs=["10.88.0.1/32"],
        ),
    )
    request = _request(client_host="10.88.0.1", host="lb.example")
    assert is_host_os_request(request) is True


def test_host_os_request_rejects_other_private_clients(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        request_locality,
        "get_settings",
        lambda: SimpleNamespace(
            firewall_trust_proxy_headers=False,
            firewall_trusted_proxy_cidrs=[],
            insecure_allow_remote_no_auth_host_cidrs=["10.88.0.1/32"],
        ),
    )
    request = _request(client_host="10.88.0.2", host="lb.example")
    assert is_host_os_request(request) is False


def test_host_os_request_accepts_localhost_host_header_with_host_network_proof(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        request_locality,
        "_insecure_allow_remote_no_auth_host_networks",
        lambda: (ip_network("10.88.0.0/24"),),
    )
    request = _request(client_host="10.88.0.176", host="localhost:2455")
    assert is_host_os_request(request) is True


def test_host_os_request_rejects_localhost_host_header_without_host_network_proof() -> None:
    request = _request(client_host="203.0.113.44", host="localhost:2455")
    assert is_host_os_request(request) is False



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel, ConfigDict, field_serializer
from pydantic.alias_generators import to_camel


class DashboardModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        ser_json_timedelta="iso8601",
    )

    @field_serializer("*", when_used="json")
    def serialize_datetime_as_utc(value, _info):
        if isinstance(value, datetime):
            if value.tzinfo is None:
                return value.isoformat() + "Z"
            return value.isoformat().replace("+00:00", "Z")
        return value



from __future__ import annotations

import random
import re

_RETRY_PATTERN = re.compile(r"(?i)try again in\s*(\d+(?:\.\d+)?)\s*(s|ms|seconds?)")
_BACKOFF_INITIAL_DELAY_MS = 200
_BACKOFF_FACTOR = 2.0
_BACKOFF_JITTER_MIN = 0.9
_BACKOFF_JITTER_MAX = 1.1


def parse_retry_after(message: str) -> float | None:
    match = _RETRY_PATTERN.search(message or "")
    if not match:
        return None
    value = float(match.group(1))
    unit = match.group(2).lower()
    if unit == "ms":
        return value / 1000
    return value


def backoff_seconds(attempt: int) -> float:
    if attempt < 1:
        attempt = 1
    exponent = _BACKOFF_FACTOR ** (attempt - 1)
    base_ms = _BACKOFF_INITIAL_DELAY_MS * exponent
    jitter = random.uniform(_BACKOFF_JITTER_MIN, _BACKOFF_JITTER_MAX)
    return (base_ms * jitter) / 1000.0



from __future__ import annotations

import asyncio
from collections import deque
from contextlib import nullcontext
from datetime import datetime, timezone
from types import SimpleNamespace
from typing import Any, cast
from unittest.mock import AsyncMock

import anyio
import pytest

from app.core.clients.proxy_websocket import UpstreamResponsesWebSocket
from app.db.models import AccountStatus
from app.modules.proxy import service as proxy_service

pytestmark = pytest.mark.unit


def _make_api_key(
    *,
    key_id: str,
    assigned_account_ids: list[str],
    account_assignment_scope_enabled: bool | None = None,
) -> proxy_service.ApiKeyData:
    return proxy_service.ApiKeyData(
        id=key_id,
        name="bridge-key",
        key_prefix="sk-bridge",
        allowed_models=None,
        enforced_model=None,
        enforced_reasoning_effort=None,
        enforced_service_tier=None,
        expires_at=None,
        is_active=True,
        created_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
        last_used_at=None,
        account_assignment_scope_enabled=(
            bool(assigned_account_ids) if account_assignment_scope_enabled is None else account_assignment_scope_enabled
        ),
        assigned_account_ids=assigned_account_ids,
    )


@pytest.mark.asyncio
async def test_get_or_create_http_bridge_session_reuses_live_local_session_without_ring_lookup(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    service = proxy_service.ProxyService(cast(Any, nullcontext()))
    key = proxy_service._HTTPBridgeSessionKey("prompt_cache_key", "bridge-key", None)
    existing = proxy_service._HTTPBridgeSession(
        key=key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        request_model="gpt-5.4-mini",
        account=cast(Any, SimpleNamespace(id="acc-1", status=AccountStatus.ACTIVE)),
        upstream=cast(UpstreamResponsesWebSocket, SimpleNamespace()),
        upstream_control=proxy_service._WebSocketUpstreamControl(),
        pending_requests=deque(),
        pending_lock=anyio.Lock(),
        response_create_gate=asyncio.Semaphore(1),
        queued_request_count=0,
        last_used_at=1.0,
        idle_ttl_seconds=120.0,
    )
    service._http_bridge_sessions[key] = existing
    monkeypatch.setattr(
        service,
        "_prune_http_bridge_sessions_locked",
        AsyncMock(),
    )
    monkeypatch.setattr(
        proxy_service,
        "get_settings",
        lambda: SimpleNamespace(http_responses_session_bridge_enabled=True),
    )

    async def _unexpected_owner_lookup(*args: object, **kwargs: object) -> str:
        raise AssertionError("live local session reuse must not hit the ring")

    monkeypatch.setattr(proxy_service, "_http_bridge_owner_instance", _unexpected_owner_lookup)
    monkeypatch.setattr(proxy_service, "_active_http_bridge_instance_ring", _unexpected_owner_lookup)

    reused = await service._get_or_create_http_bridge_session(
        key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        api_key=None,
        request_model="gpt-5.4",
        idle_ttl_seconds=120.0,
        max_sessions=8,
    )

    assert reused is existing
    assert reused.request_model == "gpt-5.4"
    assert reused.last_used_at > 1.0


@pytest.mark.asyncio
async def test_get_or_create_http_bridge_session_replaces_live_session_when_account_is_no_longer_assigned(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    service = proxy_service.ProxyService(cast(Any, nullcontext()))
    key = proxy_service._HTTPBridgeSessionKey("request", "bridge-key", "key-1")
    stale_session = proxy_service._HTTPBridgeSession(
        key=key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        request_model="gpt-5.4-mini",
        account=cast(Any, SimpleNamespace(id="acc-stale", status=AccountStatus.ACTIVE)),
        upstream=cast(UpstreamResponsesWebSocket, SimpleNamespace(close=AsyncMock())),
        upstream_control=proxy_service._WebSocketUpstreamControl(),
        pending_requests=deque(),
        pending_lock=anyio.Lock(),
        response_create_gate=asyncio.Semaphore(1),
        queued_request_count=0,
        last_used_at=1.0,
        idle_ttl_seconds=120.0,
    )
    replacement_session = proxy_service._HTTPBridgeSession(
        key=key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        request_model="gpt-5.4",
        account=cast(Any, SimpleNamespace(id="acc-fresh", status=AccountStatus.ACTIVE)),
        upstream=cast(UpstreamResponsesWebSocket, SimpleNamespace(close=AsyncMock())),
        upstream_control=proxy_service._WebSocketUpstreamControl(),
        pending_requests=deque(),
        pending_lock=anyio.Lock(),
        response_create_gate=asyncio.Semaphore(1),
        queued_request_count=0,
        last_used_at=2.0,
        idle_ttl_seconds=120.0,
    )
    service._http_bridge_sessions[key] = stale_session
    monkeypatch.setattr(service, "_prune_http_bridge_sessions_locked", AsyncMock())
    monkeypatch.setattr(
        service,
        "_create_http_bridge_session",
        AsyncMock(return_value=replacement_session),
    )
    monkeypatch.setattr(
        proxy_service,
        "get_settings",
        lambda: SimpleNamespace(http_responses_session_bridge_enabled=True),
    )
    monkeypatch.setattr(proxy_service, "_http_bridge_owner_instance", AsyncMock(return_value="instance-a"))
    monkeypatch.setattr(
        proxy_service,
        "_active_http_bridge_instance_ring",
        AsyncMock(return_value=("instance-a", ["instance-a"])),
    )
    close_session = AsyncMock()
    monkeypatch.setattr(service, "_close_http_bridge_session", close_session)

    reused = await service._get_or_create_http_bridge_session(
        key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        api_key=_make_api_key(key_id="key-1", assigned_account_ids=["acc-fresh"]),
        request_model="gpt-5.4",
        idle_ttl_seconds=120.0,
        max_sessions=8,
    )

    assert reused is replacement_session
    assert service._http_bridge_sessions[key] is replacement_session
    assert stale_session.closed is True
    assert any(call.args == (stale_session,) for call in close_session.await_args_list)


@pytest.mark.asyncio
async def test_get_or_create_http_bridge_session_replaces_live_session_when_scope_becomes_empty(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    service = proxy_service.ProxyService(cast(Any, nullcontext()))
    key = proxy_service._HTTPBridgeSessionKey("request", "bridge-key", "key-1")
    stale_session = proxy_service._HTTPBridgeSession(
        key=key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        request_model="gpt-5.4-mini",
        account=cast(Any, SimpleNamespace(id="acc-stale", status=AccountStatus.ACTIVE)),
        upstream=cast(UpstreamResponsesWebSocket, SimpleNamespace(close=AsyncMock())),
        upstream_control=proxy_service._WebSocketUpstreamControl(),
        pending_requests=deque(),
        pending_lock=anyio.Lock(),
        response_create_gate=asyncio.Semaphore(1),
        queued_request_count=0,
        last_used_at=1.0,
        idle_ttl_seconds=120.0,
    )
    replacement_session = proxy_service._HTTPBridgeSession(
        key=key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        request_model="gpt-5.4",
        account=cast(Any, SimpleNamespace(id="acc-fresh", status=AccountStatus.ACTIVE)),
        upstream=cast(UpstreamResponsesWebSocket, SimpleNamespace(close=AsyncMock())),
        upstream_control=proxy_service._WebSocketUpstreamControl(),
        pending_requests=deque(),
        pending_lock=anyio.Lock(),
        response_create_gate=asyncio.Semaphore(1),
        queued_request_count=0,
        last_used_at=2.0,
        idle_ttl_seconds=120.0,
    )
    service._http_bridge_sessions[key] = stale_session
    monkeypatch.setattr(service, "_prune_http_bridge_sessions_locked", AsyncMock())
    monkeypatch.setattr(
        service,
        "_create_http_bridge_session",
        AsyncMock(return_value=replacement_session),
    )
    monkeypatch.setattr(
        proxy_service,
        "get_settings",
        lambda: SimpleNamespace(http_responses_session_bridge_enabled=True),
    )
    monkeypatch.setattr(proxy_service, "_http_bridge_owner_instance", AsyncMock(return_value="instance-a"))
    monkeypatch.setattr(
        proxy_service,
        "_active_http_bridge_instance_ring",
        AsyncMock(return_value=("instance-a", ["instance-a"])),
    )
    close_session = AsyncMock()
    monkeypatch.setattr(service, "_close_http_bridge_session", close_session)

    reused = await service._get_or_create_http_bridge_session(
        key,
        headers={},
        affinity=proxy_service._AffinityPolicy(key="bridge-key"),
        api_key=_make_api_key(
            key_id="key-1",
            assigned_account_ids=[],
            account_assignment_scope_enabled=True,
        ),
        request_model="gpt-5.4",
        idle_ttl_seconds=120.0,
        max_sessions=8,
    )

    assert reused is replacement_session
    assert service._http_bridge_sessions[key] is replacement_session
    assert stale_session.closed is True
    assert any(call.args == (stale_session,) for call in close_session.await_args_list)
